import argparse
import sys
import struct
import time
import socket
import fcntl
import termios
import signal

from .spitool import SPIConn
from .iim42652 import IIM42652

from . import linesocket
from . import log
#from . import queue

import queue

from .timer import WTimer

DEBUG = False

maxReadBlkSize = 1024 + 512
blksz = 12

class Status:
  done = False
  stopped = False
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = [0]

status = Status()
timer = WTimer()

class LastReads:
  times = []
  nmax = 25

  def push(self, v):
    self.times.append(v)
    if len(self.times) > self.nmax:
      self.times = self.times[1:]

  def __str__(self):
    self.dtimes = [ self.times[i+1] - self.times[i] for i in range(len(self.times)-1) ]
    s = f"{self.times}\n{self.dtimes}"
    return s

lastReads = LastReads()
lastFIFO = LastReads()

def getFIFO(iim42652):
  fcnt, wmfl, ovfl = iim42652.getFIFOInfo()
  #  log.msg('FIFO status', fcnt, wmfl, ovfl)
  rcnt = int(min(fcnt,maxReadBlkSize)/blksz)
  if rcnt > 0:
    v = iim42652.getFIFOPacketData(rcnt)
    fifodata = struct.pack('B'*rcnt*blksz, *v)
    status.fcnt = fcnt
    status.maxfcnt = max(fcnt, status.maxfcnt)
    # assert(len(fifodata) == rcnt)
    # assert(len(v) == rcnt)
    # assert(rcnt % blksz == 0)
  else:
    fifodata = b''

  if ovfl:
    iim42652.resetFIFO()
    log.error("FIFO overflow: FIFO %d" % (fcnt,))
    if DEBUG:
      log.msg(f"last reads: {lastReads}")
      log.msg(f"last ffills: {lastFIFO}")
    status.nresets += 1
    status.maxfcnt = 0
  if wmfl:
    status.nwmfl += 1
#      log.error("FIFO watermark reached: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, len(fifodata)))

  return fifodata


def readFIFO(sensor, ls):
  global blksz
  timer.reset()

  blksz = sensor.getFIFOBlockSize()

  log.msg('blksz', blksz)

#  log.msg(dir(socket))
#  log.msg(dir(termios))

#  if sensor.tempmanrfreq > 0:
#    tempManInter = 1 / sensor.tempmanrfreq
#    status.curTempVals = [sensor.readTemp(),]

  smplrt = sensor.fifoSampleRate
  smpldt = 1 / smplrt

  #  que = queue.SimpleQueue(1e6)
  que = queue.SimpleQueue()

  sensor['fifoMode'] = 'Stream-to-FIFO'

  while not status.done or not que.empty():
    timer.click()
    if DEBUG:
      lastReads.push(timer.time)

    if status.done:
      if not status.stopped:
        log.msg('Stopping queue')
        sensor['fifoMode'] = 'bypass'
        fifodata = b''
        status.stopped = True
    else:
      fifodata = getFIFO(sensor)
      status.nfifodata += len(fifodata)

    if DEBUG:
      lastFIFO.push(len(fifodata))

    if len(fifodata) > 0:
      try:
        que.put(fifodata)
      except:
        pass
    elif que.empty():
      time.sleep(4e-4)

    for i in range(min(3,que.qsize())):
      try:
        item = que.get()
        ls.transmitb(item)
      except BlockingIOError as ex:
        #log.msg('socket blocked for queue')
        break
      except BrokenPipeError as ex:
        log.msg('socket closed, exit')
        status.done = True
        break

    if len(fifodata) < 300:
      if smplrt < 100:
        time.sleep(smpldt/2)
      else:
        time.sleep(smpldt * 5)

#    if status.fcnt < 1000 and sensor.tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
#      status.curTempVals = [sensor.readTemp(),]

    if timer.elapsed('info', 1):
      nwrite = 0
      # nwrite = ls.sock.getsockopt(socket.SOL_SOCKET, socket.SO_NWRITE, 8)
      log.msg("%s(% 3d): %d B, %.2f KB/s, FIFO %d (max %d), ov %d wm %d, queue %d" %
            (">" if 1e-3 * timer.last > smpldt * 40 else " ", 1e-3 * timer.last / smpldt,
             status.nfifodata, 1e-3 * status.nfifodata / timer.elaps,
             status.fcnt, status.maxfcnt, status.nresets, status.nwmfl, que.qsize()))
      sys.stdout.flush()

def ctrlCHandler(sig, frame):
  log.msg(f'Signal {sig} caught, setting stop')
  status.done = True
  return

def run():
  import os
  log.name = 'fiforeader'

  log.flush = False
#  log.level = 2

  if os.geteuid() == 0:
    res = os.setpriority(os.PRIO_PROCESS, 0, -20)
    log.msg('Am root', res)

  parser = argparse.ArgumentParser(description='FIFO reader helper process')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')
  parser.add_argument('--verbose', type=int, nargs='?', const=1, help='Verbose log output')

  parser.add_argument('--outfile', type=str, help='name of output file')

  parser.add_argument('--fifo-packet', type=int)

  args = parser.parse_args(sys.argv[1:])

  log.level = 2 if args.verbose else 1

  log.msg(args)

  signal.signal(signal.SIGINT, ctrlCHandler)
  signal.signal(signal.SIGTERM, ctrlCHandler)
  signal.signal(signal.SIGHUP, ctrlCHandler)

  spi = SPIConn(args.bus, args.dev, args.speed)
  log.msg(spi)

  iim42652 = IIM42652(spi)
  log.msg(iim42652)

  iim42652.setupFIFO(args.fifo_packet, iim42652.sampleRate)

  if not iim42652.connected:
    log.error("iim42652 device not found")
    return
  else:
    log.msg("iim42652 device connected")

  ls = linesocket.LineSocket(args.outfile)
  ls.openSocket(1)

  try:
    readFIFO(iim42652, ls)
  except KeyboardInterrupt as ex:
    log.msg('FifoReader caught Ctrl-C')
  except BaseException as ex:
    log.msg(f'FifoReader caught exception {ex}')

  log.msg(f'FifoReader read {status.nfifodata} B, {status.nfifodata/blksz} samples from FIFO')
  log.msg('FifoReader exiting')

if __name__ == "__main__":
    run()
