#! /usr/bin/env python3

from . import spitool
from . import log
from . import aaftable

from .descriptors import RegDescriptor, AliasDescriptor, RegBankDescriptor, BitDescriptor, BitfieldDescriptor, JointBitfieldDescriptor, NamedDescriptor
from .descriptors import RegInt16Descriptor, RegBankInt24Descriptor, CalculatedDescriptor, ConversionDescriptor, ScaledDescriptor
from .descriptors import DiscreteDescriptor, SignedIntDescriptor, Array3Descriptor, ArrayNDescriptor, NamedFieldDescriptor, MappedDescriptor
from .descriptors import RegMBDescriptor, Reg3DDataDescriptor, ArrayItemDescriptor, DataSubsetDescriptor8B, DataSubsetDescriptor16B
from .descriptors import FifoPkgFmt1Descriptor, FifoPkgFmt2Descriptor, FifoPkgFmt3Descriptor, FifoPkgFmt4Descriptor
from .propcache import PropCache

import time
import argparse
import math
import struct
import os
import inspect

DEBUG = False

class IIM42652:

  class Register:
    def __init__(self, spi):
      self._spi = spi
      self._cache = PropCache(self)

    class _SetUserBank:
      def __init__(self, parent, bank):
        self.bank = bank
        self.parent = parent
      def __enter__(self):
        assert(self.parent.USER_BANK == 0)
        self.parent.USER_BANK = self.bank
      def __exit__(self, exc_type, exc_value, traceback):
        self.parent.USER_BANK = 0

    def _setUserBank(self, bank):
      return IIM42652.Register._SetUserBank(self, bank)

    def _get(self, addr):
      val = self._spi.get(addr)
      return val

    def _set(self, addr, data):
      return self._spi.set(addr, data)

    def _getb(self, bank, addr):
      assert(bank != 0)
      assert(self.USER_BANK == 0)
      self.USER_BANK = bank
      val = self._spi.get(addr)
      self.USER_BANK = 0
      return val

    def _setb(self, bank, addr, data):
      assert(bank != 0)
      assert(self.USER_BANK == 0)
      self.USER_BANK = bank
      res = self._spi.set(addr, data)
      self.USER_BANK = 0
      return res

    DEVICE_CONFIG   = RegDescriptor(0x11)
    softReset       = BitDescriptor(DEVICE_CONFIG, 0)
    spiModeRaw      = BitDescriptor(DEVICE_CONFIG, 4)
    spiMode         = NamedDescriptor(spiModeRaw, {0: 'mode_0_and_3', 1: 'mode_1_and_2'})

    DRIVE_CONFIG    = RegDescriptor(0x13)
    spiSlewRateRaw  = BitfieldDescriptor(DRIVE_CONFIG, 3)
    i2cSlewRateRaw  = BitfieldDescriptor(DRIVE_CONFIG, 3, 3)
    _slewRates      = {0: '20ns-60ns',
                       1: '12ns-36ns',
                       2: '6ns-18ns',
                       3: '4ns-12ns',
                       4: '2ns-6ns',
                       5: '<2ns',
                       6: 'reserved',
                       7: 'reserved'}
    spiSlewRate     = NamedDescriptor(spiSlewRateRaw, _slewRates)
    i2cSlewRate     = NamedDescriptor(i2cSlewRateRaw, _slewRates)

    _intPolarities  = {0: 'activeLow', 1: 'activeHigh'}
    _intDriveCircuits   = {0: 'openDrain', 1: 'pushPull'}
    _intModes       = {0: 'pulsed', 1: 'latched'}
    INT_CONFIG      = RegDescriptor(0x14)
    int1PolarityRaw = BitDescriptor(INT_CONFIG, 0)
    int1Polarity    = NamedDescriptor(int1PolarityRaw, _intPolarities)
    int1DriveCircuitRaw = BitDescriptor(INT_CONFIG, 1)
    int1DriveCircuit = NamedDescriptor(int1DriveCircuitRaw, _intDriveCircuits)
    int1ModeRaw     = BitDescriptor(INT_CONFIG, 2)
    int1Mode        = NamedDescriptor(int1ModeRaw, _intModes)
    int2PolarityRaw = BitDescriptor(INT_CONFIG, 3)
    int2Polarity    = NamedDescriptor(int2PolarityRaw, _intPolarities)
    int2DriveCircuitRaw = BitDescriptor(INT_CONFIG, 4)
    int2DriveCircuit = NamedDescriptor(int2DriveCircuitRaw, _intDriveCircuits)
    int2ModeRaw     = BitDescriptor(INT_CONFIG, 5)
    int2Mode        = NamedDescriptor(int2ModeRaw, _intModes)

    FIFO_CONFIG     = RegDescriptor(0x16)
    fifoModeInt     = BitfieldDescriptor(FIFO_CONFIG, 2, 6)
    fifoMode        = NamedDescriptor(fifoModeInt, {0: 'bypass', 1: 'Stream-to-FIFO', 2: 'STOP-on-FULL', 3: 'STOP-on-FULL'})

    TEMP_DATA       = RegInt16Descriptor(0x1D)
    tempData        = CalculatedDescriptor(TEMP_DATA, lambda _, x: (x / 132.48) + 25, '°C')


    def _getGyroValue(self, x):
      return self._cache.gyroFullScale * x / (1 << 15)

    def _getAccelValue(self, x):
      return self._cache.accelFullScale * x / (1 << 15)

    def _getGyroValues(self, vals):
      return [ self._cache.gyroFullScale * vals[i] * self._gyroFactors[i] / (1 << 15) for i in self._axes ]
    def _getAccelValues(self, vals):
      return [ self._cache.accelFullScale * vals[i] * self._accelFactors[i] / (1 << 15) for i in self._axes ]

    # ACCEL_DATAX     = RegInt16Descriptor(0x1F)
    # ACCEL_DATAY     = RegInt16Descriptor(0x21)
    # ACCEL_DATAZ     = RegInt16Descriptor(0x23)
    # accelDataX      = CalculatedDescriptor(ACCEL_DATAX, getAccelValue, 'g')
    # accelDataY      = CalculatedDescriptor(ACCEL_DATAY, getAccelValue, 'g')
    # accelDataZ      = CalculatedDescriptor(ACCEL_DATAZ, getAccelValue, 'g')

    # read Accel 3D data in one go as 6B blob
    ACCEL_3D_DATA   = Reg3DDataDescriptor(0x1F)
    accelDataXInt   = ArrayItemDescriptor(ACCEL_3D_DATA, 0)
    accelDataYInt   = ArrayItemDescriptor(ACCEL_3D_DATA, 1)
    accelDataZInt   = ArrayItemDescriptor(ACCEL_3D_DATA, 2)
    accelDataX      = CalculatedDescriptor(accelDataXInt, _getAccelValue, 'g')
    accelDataY      = CalculatedDescriptor(accelDataYInt, _getAccelValue, 'g')
    accelDataZ      = CalculatedDescriptor(accelDataZInt, _getAccelValue, 'g')

    # ACCEL_DATA_BLOB = RegMBDescriptor(0x1F, 6)
    # accelDataXs     = DataSubsetDescriptor16B(ACCEL_DATA_BLOB, 0)
    # accelDataYs     = DataSubsetDescriptor16B(ACCEL_DATA_BLOB, 2)
    # accelDataZs     = DataSubsetDescriptor16B(ACCEL_DATA_BLOB, 4)
    # accelXs         = CalculatedDescriptor(accelDataXs, getAccelValue, 'g')
    # accelYs         = CalculatedDescriptor(accelDataYs, getAccelValue, 'g')
    # accelZs         = CalculatedDescriptor(accelDataZs, getAccelValue, 'g')

    accelDataInt    = Array3Descriptor([accelDataXInt, accelDataYInt, accelDataZInt])
    accelData       = Array3Descriptor([accelDataX, accelDataY, accelDataZ])


    # GYRO_DATAX      = RegInt16Descriptor(0x25)
    # GYRO_DATAY      = RegInt16Descriptor(0x27)
    # GYRO_DATAZ      = RegInt16Descriptor(0x29)

    # gyroDataX       = CalculatedDescriptor(GYRO_DATAX, getGyroValue, '°/s')
    # gyroDataY       = CalculatedDescriptor(GYRO_DATAY, getGyroValue, '°/s')
    # gyroDataZ       = CalculatedDescriptor(GYRO_DATAZ, getGyroValue, '°/s')

    # read Gyro 3D data in one go as 6B blob
    GYRO_3D_DATA   = Reg3DDataDescriptor(0x25)
    gyroDataXInt   = ArrayItemDescriptor(GYRO_3D_DATA, 0)
    gyroDataYInt   = ArrayItemDescriptor(GYRO_3D_DATA, 1)
    gyroDataZInt   = ArrayItemDescriptor(GYRO_3D_DATA, 2)
    gyroDataX      = CalculatedDescriptor(gyroDataXInt, _getGyroValue, 'g')
    gyroDataY      = CalculatedDescriptor(gyroDataYInt, _getGyroValue, 'g')
    gyroDataZ      = CalculatedDescriptor(gyroDataZInt, _getGyroValue, 'g')

    gyroDataInt    = Array3Descriptor([gyroDataXInt, gyroDataYInt, gyroDataZInt])
    gyroData       = Array3Descriptor([gyroDataX, gyroDataY, gyroDataZ])

    # read FIFO info (interrupt flags and FIFO count) in one go as 3B blob
    FIFO_INFO_BLOB  = RegMBDescriptor(0x2D, 3)
    fi_INT_STATUS   = DataSubsetDescriptor8B(FIFO_INFO_BLOB, 0)
    fi_FIFO_COUNT   = DataSubsetDescriptor16B(FIFO_INFO_BLOB, 1)
    fi_fifoFullInt  = BitDescriptor(fi_INT_STATUS, 1)
    fi_fifoThreshInt= BitDescriptor(fi_INT_STATUS, 2)

    INT_STATUS      = RegDescriptor(0x2D)
    acgReadyInt     = BitDescriptor(INT_STATUS, 0)
    fifoFullInt     = BitDescriptor(INT_STATUS, 1)
    fifoThreshInt   = BitDescriptor(INT_STATUS, 2)
    dataReadyInt    = BitDescriptor(INT_STATUS, 3)
    resetDoneInt    = BitDescriptor(INT_STATUS, 4)
    pllReadyInt     = BitDescriptor(INT_STATUS, 5)
    uiFsyncInt      = BitDescriptor(INT_STATUS, 6)

    FIFO_COUNT      = RegInt16Descriptor(0x2E)

    fifoPkgFmt1     = FifoPkgFmt1Descriptor(0x30)
    fifoPkgFmt2     = FifoPkgFmt2Descriptor(0x30)
    fifoPkgFmt3     = FifoPkgFmt3Descriptor(0x30)
    fifoPkgFmt4     = FifoPkgFmt4Descriptor(0x30)

    stepCountFFDur  = RegInt16Descriptor(0x31)
    stepCadence     = RegDescriptor(0x33)

    APEX_DATA3      = RegDescriptor(0x34)
    activityClassInt= BitfieldDescriptor(APEX_DATA3, 2)
    activityClass   = NamedDescriptor(activityClassInt, {0: 'unknown', 1: 'walk', 2: 'run', 3: 'reserved'})
    dmpIdle         = BitDescriptor(APEX_DATA3, 2)

    APEX_DATA4      = RegDescriptor(0x35)
    tapDirInt       = BitDescriptor(APEX_DATA4, 0)
    tapAxisInt      = BitfieldDescriptor(APEX_DATA4, 2, 1)
    tapNumInt       = BitfieldDescriptor(APEX_DATA4, 2, 3)
    tapDir          = NamedDescriptor(tapDirInt, {0: 'pos', 1: 'neg'})
    tapAxis         = NamedDescriptor(tapAxisInt, {0: 'X', 1: 'Y', 2: 'Z', 3: 'reserved'})
    tapNum          = NamedDescriptor(tapNumInt, {0: 'none', 1: 'single', 2: 'double', 3: 'reserved'})

    APEX_DATA5      = RegDescriptor(0x36)
    doubleTapTimingInt=BitfieldDescriptor(APEX_DATA5, 6)
    def _getDoubleTapTimingInt(self, x):
      return x * 16 / self._cache.accelODR
    doubleTapTiming = CalculatedDescriptor(doubleTapTimingInt, _getDoubleTapTimingInt, 's')

    INT_STATUS2     = RegDescriptor(0x37)
    womXInt         = BitDescriptor(INT_STATUS2, 0)
    womYInt         = BitDescriptor(INT_STATUS2, 1)
    womZInt         = BitDescriptor(INT_STATUS2, 2)
    smdInt          = BitDescriptor(INT_STATUS2, 3)
    INT_STATUS3     = RegDescriptor(0x38)
    tapDetInt       = BitDescriptor(INT_STATUS3, 0)
    ffDetInt        = BitDescriptor(INT_STATUS3, 1)
    tiltDetInt      = BitDescriptor(INT_STATUS3, 3)
    stepCntOvfInt   = BitDescriptor(INT_STATUS3, 4)
    stepDetInt      = BitDescriptor(INT_STATUS3, 5)

    womInt          = Array3Descriptor([womXInt, womYInt, womZInt])

    SIGNAL_PATH_RESET= RegDescriptor(0x48)
    fifoFlush       = BitDescriptor(SIGNAL_PATH_RESET, 1)
    tmstStrobe      = BitDescriptor(SIGNAL_PATH_RESET, 2)
    abortAndReset   = BitDescriptor(SIGNAL_PATH_RESET, 3)
    dmpMemResetEn   = BitDescriptor(SIGNAL_PATH_RESET, 5)
    dmpInitEn       = BitDescriptor(SIGNAL_PATH_RESET, 6)

    INTF_CONFIG0    = RegDescriptor(0x4C)
    fifoHoldLastDataEn= BitDescriptor(INTF_CONFIG0, 7)
    fifoCountRec    = BitDescriptor(INTF_CONFIG0, 6)
    fifoCountEndian = BitDescriptor(INTF_CONFIG0, 5)
    sensorDataEndian= BitDescriptor(INTF_CONFIG0, 4)
    uiSifsCfg       = BitfieldDescriptor(INTF_CONFIG0, 2, 0)

    PWR_MGMT0       = RegDescriptor(0x4E)
    powerAccelRaw   = BitfieldDescriptor(PWR_MGMT0, 2, 0)
    powerAccel      = NamedDescriptor(powerAccelRaw, {0: 'off', 1: 'off', 2: 'lowpower', 3: 'lownoise'})
    powerGyroRaw    = BitfieldDescriptor(PWR_MGMT0, 2, 2)
    powerGyro       = NamedDescriptor(powerGyroRaw, {0: 'off', 1: 'standby', 2: 'reserved', 3: 'lownoise'})
    powerRCIdle     = BitDescriptor(PWR_MGMT0, 4)
    disableTemp     = BitDescriptor(PWR_MGMT0, 5)

    GYRO_CONFIG0    = RegDescriptor(0x4F)
    gyroODRRaw      = BitfieldDescriptor(GYRO_CONFIG0, 4)
    gyroODR         = DiscreteDescriptor(gyroODRRaw, [None, 32e3, 16e3, 8e3, 4e3, 2e3, 1e3, 200, 100, 50, 25, 12.5, None, None, None, 500], 'Hz')
    gyroOdr         = AliasDescriptor(gyroODR)
    gyroFullScaleRaw= BitfieldDescriptor(GYRO_CONFIG0, 3, 5)
    gyroFullScale   = DiscreteDescriptor(gyroFullScaleRaw, [2000, 1000, 500, 250, 125, 62.5, 31.25, 15.625], '°/s')

    ACCEL_CONFIG0   = RegDescriptor(0x50)
    accelODRRaw     = BitfieldDescriptor(ACCEL_CONFIG0, 4)
    accelODR        = DiscreteDescriptor(accelODRRaw, [None, 32e3, 16e3, 8e3, 4e3, 2e3, 1e3, 200, 100, 50, 25, 12.5, 6.25, 3.125, 1.5625, 500], 'Hz')
    accelOdr        = AliasDescriptor(accelODR)

    accelFullScaleRaw= BitfieldDescriptor(ACCEL_CONFIG0, 3, 5)
    accelFullScale  = DiscreteDescriptor(accelFullScaleRaw, [16, 8, 4, 2, None, None, None, None], 'g')

    GYRO_CONFIG1    = RegDescriptor(0x51)
    gyroDec2M2FiltOrdInt= BitfieldDescriptor(GYRO_CONFIG1, 2)
    gyroDec2M2FiltOrd= CalculatedDescriptor(gyroDec2M2FiltOrdInt, lambda _, x: x + 1.0, '')
    gyroUIFiltOrdInt= BitfieldDescriptor(GYRO_CONFIG1, 2, 2)
    gyroUifilterOrder= ConversionDescriptor(gyroUIFiltOrdInt, lambda _, x: x + 1.0, lambda _, x: x - 1.0, '')

    tempFiltBWInt   = BitfieldDescriptor(GYRO_CONFIG1, 3, 5)
    tempFiltBW      = DiscreteDescriptor(tempFiltBWInt, [4e3, 170, 82, 40, 20, 10, 5, 5], 'Hz')
    tempFiltLat     = DiscreteDescriptor(tempFiltBWInt, [1/8, 1, 2, 4, 8, 16, 32, 32], 'ms')

    GYRO_ACCEL_CONFIG0= RegDescriptor(0x52)
    gyroUifilterBwInt = BitfieldDescriptor(GYRO_ACCEL_CONFIG0, 3)
    gyroUifilterLowLatency= BitDescriptor(GYRO_ACCEL_CONFIG0, 3)

    bw2FreqMaps = { 0: lambda ODR: ODR/2,
                    1: lambda ODR: max(400, ODR)/4,
                    2: lambda ODR: max(400, ODR)/5,
                    3: lambda ODR: max(400, ODR)/8,
                    4: lambda ODR: max(400, ODR)/10,
                    5: lambda ODR: max(400, ODR)/16,
                    6: lambda ODR: max(400, ODR)/20,
                    7: lambda ODR: max(400, ODR)/40 }
    bw2FreqMapsLL = {
                   14 - 8: lambda ODR: max(400, ODR)/4,
                   15 - 8: lambda ODR: max(200, 8*ODR)/4 }

    def _getUiFiltBWLN(self, values):
      ODR = values[0]
      bwInt = values[1]
      lowLat = values[2]
      if ODR > 1000:
        freq = ODR / 4
      else:
        mapFun = self.bw2FreqMaps[bwInt] if not lowLat else self.bw2FreqMapsLL[bwInt]
        freq = mapFun(ODR)
      return freq

    def _setGyroUIFiltBWLN(self, freq):
      ODR = self.gyroODR
      lowLat = self.gyroUifilterLowLatency
      if ODR > 1000:
        bwInt = 1
      else:
        fmaps  = self.bw2FreqMaps if not lowLat else self.bw2FreqMapsLL
        freqDPairs = [ (abs(fmaps[k](ODR) - freq), k) for k in fmaps ]
        minPair = min(freqDPairs, key=lambda x: x[0])
        bwInt = minPair[1]
      return [ODR, bwInt, lowLat]

    def _setAccelUiFiltBWLN(self, freq):
      ODR = self.accelODR
      lowLat = self.accelUifilterLowLatency
      if ODR > 1000:
        bwInt = 1
      else:
        fmaps  = self.bw2FreqMaps if not lowLat else self.bw2FreqMapsLL
        freqDPairs = [ (abs(fmaps[k](ODR) - freq), k) for k in fmaps ]
        minPair = min(freqDPairs, key=lambda x: x[0])
        bwInt = minPair[1]
      return [ODR, bwInt, lowLat]

    gyroUIFilterVals  = ArrayNDescriptor([gyroODR, gyroUifilterBwInt, gyroUifilterLowLatency])
    gyroUifilter3dBBW = ConversionDescriptor(gyroUIFilterVals, _getUiFiltBWLN, _setGyroUIFiltBWLN, 'Hz')
    gyroUifilterBw    = AliasDescriptor(gyroUifilter3dBBW)

    accelUifilterBwInt= BitfieldDescriptor(GYRO_ACCEL_CONFIG0, 3, 4)
    accelUifilterLowLatency= BitDescriptor(GYRO_ACCEL_CONFIG0, 7)
    # accelUIFilterBWLP = NamedDescriptor(accelUifilterBwInt, {1: '1xAVG', 6: '16xAVG'})

    accelUIFilterVals = ArrayNDescriptor([accelODR, accelUifilterBwInt, accelUifilterLowLatency])
    accelUifilter3dBBW= ConversionDescriptor(accelUIFilterVals, _getUiFiltBWLN, _setAccelUiFiltBWLN, 'Hz')
    accelUifilterBw   = AliasDescriptor(accelUifilter3dBBW)

    ACCEL_CONFIG1   = RegDescriptor(0x53)
    accelDec2M2FiltOrdInt= BitfieldDescriptor(ACCEL_CONFIG1, 2, 1)
    accelDec2M2FiltOrd= CalculatedDescriptor(accelDec2M2FiltOrdInt, lambda _, x: x + 1.0, '')
    accelUIFiltOrdInt= BitfieldDescriptor(ACCEL_CONFIG1, 2, 3)
    accelUifilterOrder  = ConversionDescriptor(accelUIFiltOrdInt, lambda _, x: x + 1.0, lambda _, x: x - 1.0, '')

    TMST_CONFIG1    = RegDescriptor(0x54)
    tmstEn          = BitDescriptor(TMST_CONFIG1, 0)
    tmstFsyncEn     = BitDescriptor(TMST_CONFIG1, 1)
    tmstDeltaEn     = BitDescriptor(TMST_CONFIG1, 2)
    tmstResInt      = BitDescriptor(TMST_CONFIG1, 3)
    tmstRes         = DiscreteDescriptor(tmstResInt, [1, 16], 'µs')
    tmstToRegsEn    = BitDescriptor(TMST_CONFIG1, 4)

    APEX_CONFIG0    = RegDescriptor(0x56)
    dmpODRInt       = BitfieldDescriptor(APEX_CONFIG0, 2)
    dmpODR          = MappedDescriptor(dmpODRInt, {0: 25, 1: 500, 2: 50, 3: 100}, 'Hz')
    ffEnable        = BitDescriptor(APEX_CONFIG0, 2)
    tiltEnable      = BitDescriptor(APEX_CONFIG0, 4)
    pedEnable       = BitDescriptor(APEX_CONFIG0, 5)
    tapEnable       = BitDescriptor(APEX_CONFIG0, 6)
    dmpPowerSave    = BitDescriptor(APEX_CONFIG0, 7)

    SMD_CONFIG      = RegDescriptor(0x57)
    smdModeInt      = BitfieldDescriptor(SMD_CONFIG, 2)
    smdMode         = NamedDescriptor(SMD_CONFIG, {0: 'off', 1: 'wom', 2: 'smdShort', 3: 'smdLong'})
    womModeInt      = BitDescriptor(SMD_CONFIG, 2)
    womMode         = NamedDescriptor(womModeInt, {0: 'cmpInitial', 1: 'cmpPrevious'})
    womIntModeInt   = BitDescriptor(SMD_CONFIG, 3)
    womIntMode      = NamedDescriptor(womIntModeInt, {0: 'anyAxis', 1: 'allAxes'})

    FIFO_CONFIG1    = RegDescriptor(0x5F)
    fifoAccelEn     = BitDescriptor(FIFO_CONFIG1, 0)
    fifoGyroEn      = BitDescriptor(FIFO_CONFIG1, 1)
    fifoTempEn      = BitDescriptor(FIFO_CONFIG1, 2)
    fifoTmstFsyncEn = BitDescriptor(FIFO_CONFIG1, 3)
    fifoHiresEn     = BitDescriptor(FIFO_CONFIG1, 4)
    fifoWmGtTh      = BitDescriptor(FIFO_CONFIG1, 5)
    fifoResumePartialRd= BitDescriptor(FIFO_CONFIG1, 6)

    FIFO_WM         = RegInt16Descriptor(0x60, '<')
    FIFO_CONFIG2    = RegDescriptor(0x60)
    FIFO_CONFIG3    = RegDescriptor(0x61)
    # FIFO_CONFIG3l   = BitfieldDescriptor(FIFO_CONFIG3, 4, 0)
    # FIFO_WM         = JointBitfieldDescriptor([FIFO_CONFIG2, FIFO_CONFIG3l])

    FIFO_LOST_PKT_CNT= RegInt16Descriptor(0x6C, end='<', dtype='H')

    SELF_TEST_CONFIG= RegDescriptor(0x70)
    enGyroXST       = BitDescriptor(SELF_TEST_CONFIG, 0)
    enGyroYST       = BitDescriptor(SELF_TEST_CONFIG, 1)
    enGyroZST       = BitDescriptor(SELF_TEST_CONFIG, 2)
    enAccelXST      = BitDescriptor(SELF_TEST_CONFIG, 3)
    enAccelYST      = BitDescriptor(SELF_TEST_CONFIG, 4)
    enAccelZST      = BitDescriptor(SELF_TEST_CONFIG, 5)
    accelSTPower    = BitDescriptor(SELF_TEST_CONFIG, 6)

    WHOAMI          = RegDescriptor(0x75)
    USER_BANK       = RegDescriptor(0x76)

    SENSOR_CONFIG0  = RegBankDescriptor(1, 0x3)
    xa_disable      = BitDescriptor(SENSOR_CONFIG0, 0)
    ya_disable      = BitDescriptor(SENSOR_CONFIG0, 1)
    za_disable      = BitDescriptor(SENSOR_CONFIG0, 2)
    xg_disable      = BitDescriptor(SENSOR_CONFIG0, 3)
    yg_disable      = BitDescriptor(SENSOR_CONFIG0, 4)
    zg_disable      = BitDescriptor(SENSOR_CONFIG0, 5)

    a_disable       = Array3Descriptor([xa_disable, ya_disable, za_disable])
    g_disable       = Array3Descriptor([xg_disable, yg_disable, zg_disable])

    GYRO_CONFIG_STATIC2      = RegBankDescriptor(1, 0xB)
    gyroNfilterDisable       = BitDescriptor(GYRO_CONFIG_STATIC2, 0)
    gyroAafilterDisable      = BitDescriptor(GYRO_CONFIG_STATIC2, 1)

    GYRO_CONFIG_STATIC3      = RegBankDescriptor(1, 0xC)
    gyroAAFilterDelta        = BitfieldDescriptor(GYRO_CONFIG_STATIC3, 6)

    GYRO_CONFIG_STATIC4      = RegBankDescriptor(1, 0xD)
    gyroAAFilterDeltaSqrL    = BitfieldDescriptor(GYRO_CONFIG_STATIC4, 8)
    GYRO_CONFIG_STATIC5      = RegBankDescriptor(1, 0xE)
    gyroAAFilterBitshift     = BitfieldDescriptor(GYRO_CONFIG_STATIC5, 4, 4)
    gyroAAFilterDeltaSqrH    = BitfieldDescriptor(GYRO_CONFIG_STATIC5, 4)
    gyroAAFilterDeltaSqr     = JointBitfieldDescriptor([gyroAAFilterDeltaSqrL, gyroAAFilterDeltaSqrH])

    def _getAAFilterBW(self, values):
      freq = None
      for aaf in aaftable.aaftable:
        if values == aaf[1:]:
          freq = aaf[0]
      if DEBUG:
        print(f'AAF {values} => {freq}')
      return freq

    def _setAAFilterBW(self, freq):
      adist = [ (abs(aaf[0] - freq), i) for (i,aaf) in enumerate(aaftable.aaftable) ]
      minpair = min(adist, key=lambda x: x[0])
      values = aaftable.aaftable[minpair[1]][1:]
      if DEBUG:
        print(f'AAF {freq} => {values}')
      return values

    gyroAAFilterVals         = ArrayNDescriptor([gyroAAFilterDelta, gyroAAFilterDeltaSqr, gyroAAFilterBitshift])
    gyroAAFilter3dBBW        = ConversionDescriptor(gyroAAFilterVals, _getAAFilterBW, _setAAFilterBW, 'Hz')
    gyroAafilterBw           = AliasDescriptor(gyroAAFilter3dBBW)

    GYRO_CONFIG_STATIC6      = RegBankDescriptor(1, 0xF)
    gyroXNFilterCosWZL       = BitfieldDescriptor(GYRO_CONFIG_STATIC6, 8)
    GYRO_CONFIG_STATIC7      = RegBankDescriptor(1, 0x10)
    gyroYNFilterCosWZL       = BitfieldDescriptor(GYRO_CONFIG_STATIC7, 8)
    GYRO_CONFIG_STATIC8      = RegBankDescriptor(1, 0x11)
    gyroZNFilterCosWZL       = BitfieldDescriptor(GYRO_CONFIG_STATIC8, 8)
    GYRO_CONFIG_STATIC9      = RegBankDescriptor(1, 0x12)
    gyroXNFilterCosWZH       = BitfieldDescriptor(GYRO_CONFIG_STATIC9, 1, 0)
    gyroYNFilterCosWZH       = BitfieldDescriptor(GYRO_CONFIG_STATIC9, 1, 1)
    gyroZNFilterCosWZH       = BitfieldDescriptor(GYRO_CONFIG_STATIC9, 1, 2)

    gyroXNFilterCosWZSel     = BitDescriptor(GYRO_CONFIG_STATIC9, 3)
    gyroYNFilterCosWZSel     = BitDescriptor(GYRO_CONFIG_STATIC9, 4)
    gyroZNFilterCosWZSel     = BitDescriptor(GYRO_CONFIG_STATIC9, 5)

    gyroXNFilterCosWZBin     = JointBitfieldDescriptor([gyroXNFilterCosWZL, gyroXNFilterCosWZH])
    gyroYNFilterCosWZBin     = JointBitfieldDescriptor([gyroYNFilterCosWZL, gyroYNFilterCosWZH])
    gyroZNFilterCosWZBin     = JointBitfieldDescriptor([gyroZNFilterCosWZL, gyroZNFilterCosWZH])

    gyroXNFilterCosWZ        = SignedIntDescriptor(gyroXNFilterCosWZBin)
    gyroYNFilterCosWZ        = SignedIntDescriptor(gyroYNFilterCosWZBin)
    gyroZNFilterCosWZ        = SignedIntDescriptor(gyroZNFilterCosWZBin)

    def _getGyroNotchFilterFreq(self, values):
      nf_coswz       = values[0]
      nf_coswz_sel   = values[1]
      if nf_coswz_sel == 0:
        coswz = nf_coswz / 256
      else:
        if nf_coswz > 0:
          coswz = -(nf_coswz / (8*256) - 1)
        else:
          coswz = (-nf_coswz / (8*256) - 1)
      fdes = 1e3 * math.acos(coswz)*8/(2*math.pi)
      if DEBUG:
        print(f'gyro notch filter config: nf_coswz={nf_coswz}, nf_coswz_sel={nf_coswz_sel} => coswz={coswz}, f={fdes} Hz')
      return fdes

    def _setGyroNotchFilterFreq(self, freq):
      coswz = math.cos(2*math.pi*1e-3*freq/8)
      if DEBUG:
        print(f'gyro notch filter: f={freq} Hz => coswz={coswz}')
      if abs(coswz) <= 0.875:
        nf_coswz = round(coswz*256)
        nf_coswz_sel = 0
      else:
        nf_coswz_sel = 1
        if coswz > 0:
          nf_coswz = round(8*(1-coswz)*256)
        else:
          nf_coswz = round(-8*(1+coswz)*256)
      if DEBUG:
        print(f'gyro notch filter: f={freq} Hz => coswz={coswz}, nf_coswz={nf_coswz}, nf_coswz_sel={nf_coswz_sel}')
      values = (nf_coswz, nf_coswz_sel)
      return values

    gyroXNFilterConf         = ArrayNDescriptor([gyroXNFilterCosWZ, gyroXNFilterCosWZSel])
    gyroXNFilterFreq         = ConversionDescriptor(gyroXNFilterConf, _getGyroNotchFilterFreq, _setGyroNotchFilterFreq, 'Hz')
    gyroYNFilterConf         = ArrayNDescriptor([gyroYNFilterCosWZ, gyroYNFilterCosWZSel])
    gyroYNFilterFreq         = ConversionDescriptor(gyroYNFilterConf, _getGyroNotchFilterFreq, _setGyroNotchFilterFreq, 'Hz')
    gyroZNFilterConf         = ArrayNDescriptor([gyroZNFilterCosWZ, gyroZNFilterCosWZSel])
    gyroZNFilterFreq         = ConversionDescriptor(gyroZNFilterConf, _getGyroNotchFilterFreq, _setGyroNotchFilterFreq, 'Hz')

    gyroNfilterCenterX       = AliasDescriptor(gyroXNFilterFreq)
    gyroNfilterCenterY       = AliasDescriptor(gyroYNFilterFreq)
    gyroNfilterCenterZ       = AliasDescriptor(gyroZNFilterFreq)

    GYRO_CONFIG_STATIC10     = RegBankDescriptor(1, 0x13)
    gyroHPFOrderInd          = BitfieldDescriptor(GYRO_CONFIG_STATIC10, 1)
    gyroHPFBwInd             = BitfieldDescriptor(GYRO_CONFIG_STATIC10, 3, 1)
    gyroNFBwSel              = BitfieldDescriptor(GYRO_CONFIG_STATIC10, 3, 4)
    gyroNF3dBBW              = DiscreteDescriptor(gyroNFBwSel, [1449, 680, 329, 162, 80, 40, 20, 10], 'Hz')
    gyroNfilterBw            = AliasDescriptor(gyroNF3dBBW)

    XG_ST_DATA               = RegBankDescriptor(1, 0x5f)
    YG_ST_DATA               = RegBankDescriptor(1, 0x60)
    ZG_ST_DATA               = RegBankDescriptor(1, 0x61)
    gyroSelfTestData         = Array3Descriptor([XG_ST_DATA, YG_ST_DATA, ZG_ST_DATA])

    TMST_VALUE               = RegBankInt24Descriptor(1, 0x62)


    ACCEL_CONFIG_STATIC2     = RegBankDescriptor(2, 0x3)
    accelAafilterDisable     = BitDescriptor(ACCEL_CONFIG_STATIC2, 0)
    accelAAFilterDelta       = BitfieldDescriptor(ACCEL_CONFIG_STATIC2, 6, 1)

    ACCEL_CONFIG_STATIC3     = RegBankDescriptor(2, 0x4)
    accelAAFilterDeltaSqrL   = BitfieldDescriptor(ACCEL_CONFIG_STATIC3, 8)
    ACCEL_CONFIG_STATIC4     = RegBankDescriptor(2, 0x5)
    accelAAFilterBitshift    = BitfieldDescriptor(ACCEL_CONFIG_STATIC4, 4, 4)
    accelAAFilterDeltaSqrH   = BitfieldDescriptor(ACCEL_CONFIG_STATIC4, 4)
    accelAAFilterDeltaSqr    = JointBitfieldDescriptor([accelAAFilterDeltaSqrL, accelAAFilterDeltaSqrH])

    accelAAFilterVals        = ArrayNDescriptor([accelAAFilterDelta, accelAAFilterDeltaSqr, accelAAFilterBitshift])
    accelAAFilter3dBBW       = ConversionDescriptor(accelAAFilterVals, _getAAFilterBW, _setAAFilterBW, 'Hz')
    accelAafilterBw          = AliasDescriptor(accelAAFilter3dBBW)

    XA_ST_DATA               = RegBankDescriptor(2, 0x3B)
    YA_ST_DATA               = RegBankDescriptor(2, 0x3C)
    ZA_ST_DATA               = RegBankDescriptor(2, 0x3D)
    accelSelfTestData        = Array3Descriptor([XA_ST_DATA, YA_ST_DATA, ZA_ST_DATA])


    FDR_CONFIG               = RegBankDescriptor(4, 0x9)
    fdrSelInt                = BitfieldDescriptor(FDR_CONFIG, 7)
    fdrSel                   = ConversionDescriptor(fdrSelInt, lambda _, x: x + 1.0, lambda _, x: x - 1.0, '')

    APEX_CONFIG1             = RegBankDescriptor(4, 0x40)
    dmpPowerSaveTimeSelInt   = BitfieldDescriptor(APEX_CONFIG1, 4)
    dmpPowerSaveTimeSel      = ScaledDescriptor(dmpPowerSaveTimeSelInt, 4.0, 's')
    lowEnergyAmpThreshSel    = BitfieldDescriptor(APEX_CONFIG1, 4, 4)

    APEX_CONFIG2             = RegBankDescriptor(4, 0x41)
    pedStepCntThreshSel      = BitfieldDescriptor(APEX_CONFIG2, 4)
    pedAmplThreshSel         = BitfieldDescriptor(APEX_CONFIG2, 4, 4)

    APEX_CONFIG3             = RegBankDescriptor(4, 0x42)
    pedHiEnThreshSel         = BitfieldDescriptor(APEX_CONFIG3, 2)
    pedSBTimerThreshSel      = BitfieldDescriptor(APEX_CONFIG3, 3, 2)
    pedStepDetThreshSel      = BitfieldDescriptor(APEX_CONFIG3, 3, 5)

    APEX_CONFIG4             = RegBankDescriptor(4, 0x43)
    highGPeakThreshHystSelInt= BitfieldDescriptor(APEX_CONFIG4, 3)
    highGPeakThreshHystSel   = DiscreteDescriptor(highGPeakThreshHystSelInt, [31,63,94,125,156,188,219,250], 'mgee')
    lowGPeakThreshHystSelInt = BitfieldDescriptor(APEX_CONFIG4, 3, 3)
    lowGPeakThreshHystSel    = DiscreteDescriptor(lowGPeakThreshHystSelInt, [31,63,94,125,156,188,219,250], 'mgee')
    tiltWaitTimeSel          = BitfieldDescriptor(APEX_CONFIG4, 2, 6)
    tiltWaitTime             = ScaledDescriptor(tiltWaitTimeSel, 2.0, 's')

    APEX_CONFIG7             = RegBankDescriptor(4, 0x46)
    tapMaxPeakTol            = BitfieldDescriptor(APEX_CONFIG7, 2)
    tapMinJerkThresh         = BitfieldDescriptor(APEX_CONFIG7, 6, 2)

    APEX_CONFIG8             = RegBankDescriptor(4, 0x47)
    tapTMin                  = BitfieldDescriptor(APEX_CONFIG8, 3)
    tapTAvg                  = BitfieldDescriptor(APEX_CONFIG8, 2, 3)
    tapTMax                  = BitfieldDescriptor(APEX_CONFIG8, 2, 5)

    APEX_CONFIG8             = RegBankDescriptor(4, 0x48)
    sensitivityModeInt       = BitDescriptor(APEX_CONFIG8, 0)
    sensitivityMode          = NamedDescriptor(sensitivityModeInt, {0: 'lowpow/hiper', 1: 'lowpow/slowwalk'})

    ACCEL_WOM_X_THR          = RegBankDescriptor(4, 0x4A)
    ACCEL_WOM_Y_THR          = RegBankDescriptor(4, 0x4B)
    ACCEL_WOM_Z_THR          = RegBankDescriptor(4, 0x4C)
    accelWomXThr             = ScaledDescriptor(ACCEL_WOM_X_THR, 1/256, 'g')
    accelWomYThr             = ScaledDescriptor(ACCEL_WOM_Y_THR, 1/256, 'g')
    accelWomZThr             = ScaledDescriptor(ACCEL_WOM_Z_THR, 1/256, 'g')
    accelWomThr              = Array3Descriptor([accelWomXThr, accelWomYThr, accelWomZThr])

    INT_SOURCE6              = RegBankDescriptor(4, 0x4D)
    INT_SOURCE7              = RegBankDescriptor(4, 0x4F)

    OFFSET_USER0             = RegBankDescriptor(4, 0x77)
    OFFSET_USER1             = RegBankDescriptor(4, 0x78)
    OFFSET_USER2             = RegBankDescriptor(4, 0x79)
    OFFSET_USER3             = RegBankDescriptor(4, 0x7A)
    OFFSET_USER4             = RegBankDescriptor(4, 0x7B)
    OFFSET_USER5             = RegBankDescriptor(4, 0x7C)
    OFFSET_USER6             = RegBankDescriptor(4, 0x7D)
    OFFSET_USER7             = RegBankDescriptor(4, 0x7E)
    OFFSET_USER8             = RegBankDescriptor(4, 0x7F)

    gyroXOffsetH             = BitfieldDescriptor(OFFSET_USER1, 4)
    gyroYOffsetH             = BitfieldDescriptor(OFFSET_USER1, 4, 4)
    gyroZOffsetH             = BitfieldDescriptor(OFFSET_USER4, 4)
    accelXOffsetH            = BitfieldDescriptor(OFFSET_USER4, 4, 4)
    accelYOffsetH            = BitfieldDescriptor(OFFSET_USER7, 4)
    accelZOffsetH            = BitfieldDescriptor(OFFSET_USER7, 4, 4)

    gyroXOffsetBin           = JointBitfieldDescriptor([OFFSET_USER0, gyroXOffsetH])
    gyroYOffsetBin           = JointBitfieldDescriptor([OFFSET_USER2, gyroYOffsetH])
    gyroZOffsetBin           = JointBitfieldDescriptor([OFFSET_USER3, gyroZOffsetH])

    gyroXOffsetInt           = SignedIntDescriptor(gyroXOffsetBin)
    gyroYOffsetInt           = SignedIntDescriptor(gyroYOffsetBin)
    gyroZOffsetInt           = SignedIntDescriptor(gyroZOffsetBin)

    gyroOffsetX              = ScaledDescriptor(gyroXOffsetInt, 1.0/32, '°/s')
    gyroOffsetY              = ScaledDescriptor(gyroYOffsetInt, 1.0/32, '°/s')
    gyroOffsetZ              = ScaledDescriptor(gyroZOffsetInt, 1.0/32, '°/s')

    gyroOffset               = Array3Descriptor([gyroOffsetX, gyroOffsetY, gyroOffsetZ])

    accelXOffsetBin          = JointBitfieldDescriptor([OFFSET_USER5, accelXOffsetH])
    accelYOffsetBin          = JointBitfieldDescriptor([OFFSET_USER6, accelYOffsetH])
    accelZOffsetBin          = JointBitfieldDescriptor([OFFSET_USER8, accelZOffsetH])

    accelXOffsetInt          = SignedIntDescriptor(accelXOffsetBin)
    accelYOffsetInt          = SignedIntDescriptor(accelYOffsetBin)
    accelZOffsetInt          = SignedIntDescriptor(accelZOffsetBin)

    accelOffsetX             = ScaledDescriptor(accelXOffsetInt, 0.5e-3, 'g')
    accelOffsetY             = ScaledDescriptor(accelYOffsetInt, 0.5e-3, 'g')
    accelOffsetZ             = ScaledDescriptor(accelZOffsetInt, 0.5e-3, 'g')

    accelOffset              = Array3Descriptor([accelOffsetX, accelOffsetY, accelOffsetZ])


  tempmanrfreq = 100

  def getRegisters(self):
    dnames = [ v for v in dir(self.register) if v[0] in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' ]
    regs = { v: self[v] for v in dnames }
    return regs

  def getPropertyValues(self):
    dnames = [ v for v in dir(self.register) if not v.startswith('_') ]
    regs = { v: self[v] for v in dnames }
    return regs

  def __init__(self, spi):
    self.register = IIM42652.Register(spi)
    self.reginfo = [(getattr(getattr(IIM42652.Register, f), 'bank', 0), getattr(getattr(IIM42652.Register, f), 'reg', 0), f) \
                    for f in dir(self.register) if not f.startswith('_')]
    self.reginfo.sort()
    self.regkeys = [f[2] for f in self.reginfo]
    self.spi = spi

  @classmethod
  def underScoreToCamel(self, key):
    r = ''
    upperN = False
    for c in key:
      if c == '_':
        upperN = True
      elif upperN:
        upperN = False
        r += c.upper()
      else:
        r += c
    return r

  def setOptions(self, args):
    for i in args:
      cn = self.underScoreToCamel(i)
      if cn in self.regkeys:
        if args[i] is not None:
          log.msg(f'Setting option {cn}={args[i]}')
          self[cn] = args[i]
          log.msg(f'Setting option {cn}={args[i]} => {self[cn]}')
      else:
        if DEBUG:
          print(f'Error: Option {i}/{cn} is not a setting')
    accelScales = [ args['accel_factor_x'], args['accel_factor_y'], args['accel_factor_z'] ]
    if args['accel_g'] != 1:
      normAccel = 9.80665
      accelScales = [ v * normAccel for v in accelScales ]
    self.register._accelFactors = accelScales
    self.register._gyroFactors = [ args['gyro_factor_x'], args['gyro_factor_y'], args['gyro_factor_z'] ]
    assert(len(args['axes']) == 3)
    assert(0 in args['axes'])
    assert(1 in args['axes'])
    assert(2 in args['axes'])
    self.register._axes = args['axes']
    log.msg(f'Axes configuration: {self.register._axes}')
    log.msg(f'Accel factors: {self.register._accelFactors}')
    log.msg(f'Gyro factors: {self.register._gyroFactors}')

  @classmethod
  def getProperty(self, name):
    prop = getattr(IIM42652.Register, name)
    return prop

  def __getitem__(self, key):
    if key not in self.regkeys:
      raise KeyError
    return getattr(self.register, key)

  def __setitem__(self, key, value):
    if key not in self.regkeys:
      raise KeyError
    return setattr(self.register, key, value)

  def __iter__(self):
    self.location = 0
    return self

  def __next__(self):
    if self.location == len(self.regkeys):
      raise StopIteration
    value = self.regkeys[self.location]
    self.location += 1
    return value

  def __repr__(self):
    s = 'IIM42652({'
    for key in self:
      s += f'{key:>24s}: {self[key]},\n'
    s += '})'
    return s

  def __str__(self):
    s = "IIM42652 model\n"
    return s

  def reset(self):
    if not self.connected:
      log.error("iim42652.reset: device not found")
      return
    done = False
    self.register.softReset = 1
    for i in range(100):
      time.sleep(1e-4)
      if self.register.resetDoneInt:
        done = True
        print(f"device reset after {(i+1)*1e-4} s")
        break
    if not done:
      log.error("iim42652.reset: device reset failed")

  @property
  def connected(self):
    valWhoami = 0x6f
    return self.register.WHOAMI == valWhoami

  def wait4data(self, timeout=1000):
    tend = time.time() + timeout*1e-3
    while time.time() < tend:
      d = self.register.dataReadyInt
      if d:
        #print("data ready", d)
        break
      time.sleep(5e-4)

  def setup(self):
    iim42652 = self

  def resetFIFO(self):
    self.register.fifoModeInt = 0
    self.register.fifoModeInt = 1

  fifoMode = 0
  def stopFIFO(self):
    iim42652 = self
    self.fifoMode = 0
    iim42652['fifoModeInt'] = 0
    iim42652['FIFO_CONFIG1'] = 0

  def setupFIFO(self, mode, odr):
    if mode == 0:
      self.stopFIFO()
      return
    log.msg(f'setup FIFO: {mode}, {odr}')
    self.fifoMode = mode
    iim42652 = self
    iim42652['fifoHoldLastDataEn'] = 1
    iim42652['FIFO_WM'] = 1600
    iim42652['FIFO_CONFIG1'] = 0
    iim42652['fifoWmGtTh'] = True
    iim42652['fifoResumePartialRd'] = True
    iim42652['fifoTempEn'] = True
    iim42652['accelODR'] = odr
    iim42652['gyroODR'] = odr
    if mode == 1:
      iim42652['powerAccel'] = 'lownoise'
      iim42652['fifoAccelEn'] = True
    elif mode == 2:
      iim42652['powerGyro'] = 'lownoise'
      iim42652['fifoGyroEn'] = True
    elif mode == 3:
      iim42652['powerAccel'] = 'lownoise'
      iim42652['powerGyro'] = 'lownoise'
      iim42652['fifoAccelEn'] = True
      iim42652['fifoGyroEn'] = True
    elif mode == 4:
      iim42652['powerAccel'] = 'lownoise'
      iim42652['powerGyro'] = 'lownoise'
      iim42652['fifoHiresEn'] = True
    else:
      print(f'Invalid FIFO mode {mode}')

    self.FIFOPacketPropName = self.getFIFOPacketPropName()
    self.FIFOPacketProp = self.getFIFOPacketProp()
    self.FIFOInfoProp   = self.getFIFOInfoProp()
    self.FIFOBlockSize  = self.getFIFOBlockSize()
    log.msg(f'FIFO setup: packet {self.FIFOPacketPropName}, {self.FIFOBlockSize} B')

  def getFIFOPacketPropName(self):
    fifoPacketPName = f'fifoPkgFmt{self.fifoMode}'
    return fifoPacketPName

  def getFIFOPacketProp(self):
    fifoPacketProp = self.getProperty(self.getFIFOPacketPropName())
    return fifoPacketProp

  def getFIFOBlockSize(self):
    return self.getFIFOPacketProp().length

  FIFOInfoPropName = 'FIFO_INFO_BLOB'
  def getFIFOInfoProp(self):
    fifoInfoProp = self.getProperty(self.FIFOInfoPropName)
    return fifoInfoProp

  def getFIFOInfo(self):
    data = self[self.FIFOInfoPropName]
    f = self.FIFOInfoProp.splitr(None, data)
    return ( f[v] for v in ['fi_FIFO_COUNT', 'fi_fifoThreshInt', 'fi_fifoFullInt'] )

  def getFIFOPacket(self):
    return self[self.FIFOPacketPropName]

  def getFIFOPacketValues(self):
    return self[self.FIFOPacketPropName]['values']

  def getFIFOPacketData(self, n=1):
    return self.spi.getbytes(self.FIFOPacketProp.source, self.FIFOPacketProp.length*n)

  def setupGyroNotchFilter(self, fdes=2e3):
    assert(fdes >= 1000)
    assert(fdes <= 3000)
    coswz = math.cos(2*math.pi*fdes/8)
    print(f'gyro notch filter: f={fdes} Hz => coswz={coswz}')
    if abs(coswz) <= 0.875:
      nf_coswz = round(coswz*256)
      nf_coswz_sel = 0
    else:
      nf_coswz_sel = 1
      if coswz > 0:
        nf_coswz = round(8*(1-coswz)*256)
      else:
        nf_coswz = round(-8*(1+coswz)*256)
    print(f'gyro notch filter: f={fdes} Hz => nf_coswz={nf_coswz}, nf_coswz_sel={nf_coswz_sel}')

    self['gyroXNFilterCosWZ'] = nf_coswz
    self['gyroYNFilterCosWZ'] = nf_coswz
    self['gyroZNFilterCosWZ'] = nf_coswz

    self['gyroXNFilterCosWZSel'] = nf_coswz_sel
    self['gyroYNFilterCosWZSel'] = nf_coswz_sel
    self['gyroZNFilterCosWZSel'] = nf_coswz_sel

  def getGyroNotchFilter(self, axis='X'):
    nf_coswz_sel   = self[f'gyro{axis}NFilterCosWZSel']
    nf_coswz       = self[f'gyro{axis}NFilterCosWZ']
    if nf_coswz_sel == 0:
      coswz = coswz / 256
    else:
      coswz = -(nf_coswz / (8*256) - 1)
      fdes = math.acos(coswz)
    print(f'gyro notch filter config {axis}: nf_coswz={nf_coswz}, nf_coswz_sel={nf_coswz_sel} => f={fdes} Hz')

  def accelSelfTest(self):
    s = self
    s['powerAccel'] = 'lownoise'
    s['accelFullScale'] = 16
    s['accelODR'] = 1000
    time.sleep(50e-3)

    aResp1i = s['accelDataInt']
    aResp1 = s['accelData']

    s['SELF_TEST_CONFIG'] = 0b1111000
    time.sleep(50e-3)

    aResp2 = s['accelData']
    aResp2i = s['accelDataInt']
    s['SELF_TEST_CONFIG'] = 0

    aTest  = s['accelSelfTestData']
    astResp = [ aResp2i[i] - aResp1i[i] for i in range(3) ]
    astRespErr = [ abs(1e-1 * astResp[i] - aTest[i]) for i in range(3) ]
    astRespRelErr = [ 1e2 * astRespErr[i] / abs(aResp2i[i]) for i in range(3) ]

    passed = not any([ astRespRelErr[i] > 5 for i in range(3) ])

    return (passed, astResp, astRespErr, astRespRelErr)

  def gyroSelfTest(self):
    s = self
    s['powerGyro'] = 'lownoise'
    s['gyroFullScale'] = 2000
    s['gyroODR'] = 1000
    time.sleep(50e-3)

    gResp1i = s['gyroDataInt']
    gResp1 = s['gyroData']

    s['SELF_TEST_CONFIG'] = 0b111
    time.sleep(50e-3)

    gResp2 = s['gyroData']
    gResp2i = s['gyroDataInt']
    s['SELF_TEST_CONFIG'] = 0

    gTest  = s['gyroSelfTestData']

    gstResp = [ gResp2i[i] - gResp1i[i] for i in range(3) ]
    gstRespErr = [ abs(1e-1 * gstResp[i] - gTest[i]) for i in range(3) ]
    gstRespRelErr = [ 1e2 * gstRespErr[i] / abs(gResp2i[i]) for i in range(3) ]

    passed = not any([ gstRespRelErr[i] > 5 for i in range(3) ])

    return (passed, gstResp, gstRespErr, gstRespRelErr)

  @property
  def sampleRate(self):
    if self['powerAccel'] != 'off' and self['powerGyro'] != 'off':
      assert(self['accelODR'] == self['accelODR'])
      return self['accelODR']
    elif self['powerAccel'] != 'off':
      return self['accelODR']
    elif self['powerGyro'] != 'off':
      return self['gyroODR']

  @property
  def fifoSampleRate(self):
    return self.sampleRate / self['fdrSel']

def parseargs(cmdargs=None):
  import argparse
  import configparser

  conf_parser = argparse.ArgumentParser(add_help=False)
  conf_parser.add_argument("-c", "--conf_file", default=["/etc/iim42652.conf", "iim42652.conf"],
                           help="Specify config file", metavar="FILE")
  conf_parser.add_argument("-C", "--conf_section", default="iimsensor",
                           help="Specify config file section", metavar="NAME")
  args, remaining_argv = conf_parser.parse_known_args(cmdargs)

  def_dev = 0
  try:
    def_dev_env = os.getenv('IIM42652_SPI_DEV')
    if def_dev_env:
      def_dev = int(def_dev_env)
  except BaseException as e:
    print(e)
    pass

  defaults = {
    'debug': 0,
    'trace': 0,

    'bus': 0,
    'dev': def_dev,
    'speed': int(2e6),       # Hz

    'maxreads': -1,
    'maxreadt': -1,
    'raw_values': 0,
    'ilff': 0,
    'output': 'file',
    'nclients': 1,

    'outfile': 'out.csv',
    'downsample_outfile': '',
    'downsample_rate': 100,
    'events_outfile': '',
    'precision': 8,
    'format': 'eng',

    'axes': [0, 1, 2],

    'accel_factor_x': 1,
    'accel_factor_y': 1,
    'accel_factor_z': -1,
    'accel_cycle': 0,
    'accel_disable': 0,
    'accel_freq': 1125,      # Hz
    'accel_scale': 4,        # g
    'accel_use_filter': 1,
    'accel_uifilter_bw_int': 0,
    'accel_avg': 1,
    'accel_g': 1,

    'gyro_factor_x': 1,
    'gyro_factor_y': 1,
    'gyro_factor_z': -1,
    'gyro_cycle': 0,
    'gyro_disable': 0,
    'gyro_freq': 1125,       # Hz
    'gyro_full_scale': 2000,      # deg/s (dps)
    'gyro_use_filter': 0,
    'gyro_uifilter_bw_int': 0,
    'gyro_avg': 1,

    'temp_disable': 1,
    'temp_filter_bw': 8000,

    'temp_ext_enable': [],
    'temp_ext_b': 3500,

    'tempmanrfreq': 10,
    'magmanrfreq': 100,

    'odr': 1000,
    'accel_odr': 1000,
    'gyro_odr': 1000,
    'fifo_packet': 4,
    'fdr_sel': 1,

  }
  for key in defaults:
    if not key.islower():
      print('default dict keys must be lower case:', key)
      exit(1)
  if args.conf_file:
    config = configparser.ConfigParser(defaults=defaults)
#    print("Reading configuration file from %s" % (args.conf_file,))
    cread = config.read(args.conf_file)
    print("Read configuration files: %s" % (cread,))
    if len(cread):
      defaults = dict(config.items(args.conf_section))
      # # List values need conversion str -> list
      for key in ['temp_ext_enable', 'axes']:
        if defaults[key].startswith('['):
          print(key, defaults[key], '=>', eval(defaults[key]))
          defaults[key] = eval(defaults[key])
      print('Loading configuration defaults from %s/[%s]' % (cread, args.conf_section))

  parser = argparse.ArgumentParser(description='Read IIM42652 sensor',
                                   parents=[conf_parser],
                                   formatter_class=argparse.RawDescriptionHelpFormatter)

  parser.add_argument('--debug', type=int, nargs='?', const=1, help='Debug')
  parser.add_argument('--trace', type=int, nargs='?', const=1, help='Trace')
  parser.add_argument('--verbose', type=int, nargs='?', const=1, help='Verbose log output')
  parser.add_argument('--bus', type=int, help='SPI bus')
  parser.add_argument('--dev', type=int, help='SPI device')
  parser.add_argument('--speed', type=int, help='SPI speed')
  parser.add_argument('--reset', type=int, nargs='?', const=1, help='Reset device')

  parser.add_argument('--maxreads', type=int, help='number of reads')
  parser.add_argument('--maxreadt', type=float, help='time limit of reading')
  parser.add_argument('--raw-values', dest='raw_values', type=int, help='output raw 16 bit values')
  parser.add_argument('--rel-times', dest='rel_times', type=int, help='output relative time stamps')
  parser.add_argument('--ilff', type=int, nargs='?', const=1, help='Output using ILFF')
#  parser.add_argument('--ilff', action='store_true', help='Output using ILFF')
  parser.add_argument('--output', type=str, help='Output mode: lstofile, socket, file')
  parser.add_argument('--nclients', type=int, help='Number of socket listeners')

  parser.add_argument('--odr', '-r', type=float, help='Output data rate')
  parser.add_argument('--accel-odr', type=float, help='Output data rate')
  parser.add_argument('--gyro-odr',  type=float, help='Output data rate')
  parser.add_argument('--fifo-packet', '-f', type=int, help='FIFO Packet Mode (1-4)')
  parser.add_argument('--fdr-sel', type=int, help='FIFO Packet Decimation (1-127)')

  parser.add_argument('--axes',                      type=int, nargs='*', help='Axis reorder')

  parser.add_argument('--gyro-factor-x',             type=float, help='Factor for Gyro X axis')
  parser.add_argument('--gyro-factor-y',             type=float, help='Factor for Gyro Y axis')
  parser.add_argument('--gyro-factor-z',             type=float, help='Factor for Gyro Z axis')

  parser.add_argument('--gyro-full-scale',            type=float, help='Gyro scale: 250, 500, 1000, 2000')

  parser.add_argument('--gyro-nfilter-bw',            type=float)
  parser.add_argument('--gyro-nfilter-center-x',      type=float)
  parser.add_argument('--gyro-nfilter-center-y',      type=float)
  parser.add_argument('--gyro-nfilter-center-z',      type=float)
  parser.add_argument('--gyro-nfilter-disable',       type=float)

  parser.add_argument('--gyro-aafilter-bw',           type=float)
  parser.add_argument('--gyro-aafilter-disable',      type=float)

  parser.add_argument('--gyro-offset-x',              type=float)
  parser.add_argument('--gyro-offset-y',              type=float)
  parser.add_argument('--gyro-offset-z',              type=float)

  parser.add_argument('--gyro-uifilter-bw-int',       type=int)
  parser.add_argument('--gyro-uifilter-order',        type=int)
  parser.add_argument('--gyro-uifilter-low-latency',  type=int, nargs='?', const=1)


  parser.add_argument('--accel-factor-x',             type=float, help='Factor for Accel X axis')
  parser.add_argument('--accel-factor-y',             type=float, help='Factor for Accel Y axis')
  parser.add_argument('--accel-factor-z',             type=float, help='Factor for Accel Z axis')

  parser.add_argument('--accel-full-scale',           type=float, help='Accel scale: 2, 4, 8, 16')
  parser.add_argument('--accel-g',                    type=int,   help='Accel unit: 0: m/s, 1: g')

  parser.add_argument('--accel-aafilter-bw',          type=float)
  parser.add_argument('--accel-aafilter-disable',     type=float)

  parser.add_argument('--accel-offset-x',             type=float)
  parser.add_argument('--accel-offset-y',             type=float)
  parser.add_argument('--accel-offset-z',             type=float)

  parser.add_argument('--accel-uifilter-bw-int',      type=int)
  parser.add_argument('--accel-uifilter-order',       type=int)
  parser.add_argument('--accel-uifilter-low-latency', type=int, nargs='?', const=1)


  parser.add_argument('--temp-disable',   dest='temp_disable',   type=int, help='Disable Temperature')
  parser.add_argument('--temp-filter-bw',dest='temp_filter_bw',type=float,   help='Temp filter BW: 8.3, 17.0, ..., 499 Hz')

  parser.add_argument('--temp-ext-enable', type=int, nargs='*', help='Temp manual retrieval freq (Hz)')
  parser.add_argument('--temp-ext-freq', dest='tempmanrfreq',type=float,   help='Temp manual retrieval freq (Hz)')
  parser.add_argument('--temp-ext-b', type=float,  help='Temp ext sensor param B')

  parser.add_argument('--mag-freq', dest='magmanrfreq',type=float,   help='Magnet. manual retrieval freq (Hz)')

  parser.add_argument('--outfile', '-o', type=str, help='Output CSV file')
  parser.add_argument('--downsample-outfile', type=str, help='Downsample CSV output file')
  parser.add_argument('--downsample-rate', type=int, help='Downsample CSV output with rate')
  parser.add_argument('--events-outfile', type=str, help='Events CSV output file')
  parser.add_argument('--precision', '-p', type=int, help='CSV precision')
  parser.add_argument('--format', type=str, help='CSV value format (fixed or adaptive)')

  parser.set_defaults(**defaults)

  args = parser.parse_args(remaining_argv)

  return args

def getIIM42652(args=None, reset=True, init=True):
  if args is None:
    args = parseargs([])
  if type(args) == type({}):
    nargs = parseargs([])
    for v in args:
      setattr(nargs, v, args[v])
    args = nargs
  spi = spitool.SPIConn(args.bus, args.dev, args.speed)
  iim42652 = IIM42652(spi)
  if reset:
    iim42652.reset()
  if init:
    iim42652.setup()
    iim42652.setOptions(vars(args))
  return iim42652

def getIIM42652Cmdline(init=True):
  args = parseargs()
  if args.trace:
    print(args)
    IIM42652.DEBUG = True
    spitool.DEBUG = True
  reset = args.reset if args.reset is not None else True
  return (getIIM42652(args, reset, init), args)
