#! /usr/bin/env python3

import time
import argparse
import math
import struct
import os
import sys
import subprocess

import ilff

from .iim42652 import getIIM42652Cmdline

from . import table
from . import linesocket
from . import log

def run():
  log.name = 'iim42652_fifo'
  import locale
  locale.setlocale(locale.LC_ALL, '')

  (iim42652, args) = getIIM42652Cmdline()

  nfifodata = 0

  fifoInfoPropName = 'FIFO_INFO_BLOB'
  fifoInfoProp = iim42652.getProperty(fifoInfoPropName)

  def getFInfo():
    data = iim42652[fifoInfoPropName]
    f = fifoInfoProp.splitr(None, data)
    return ( f[v] for v in ['fi_FIFO_COUNT', 'fi_fifoThreshInt', 'fi_fifoFullInt'] )

  rawWords = []
  nreads = 0
  nsamples = 0
  process = True

  fifoMode = args.fifo_mode
  iim42652.stopFIFO()
  iim42652.setupFIFO(fifoMode, args.odr)

  iim42652['accelFullScale'] = 2
  iim42652['gyroFullScale'] = 250

  fifoPacketPName = f'fifoPkgFmt{fifoMode}'
  fifoPacketProp = iim42652.getProperty(fifoPacketPName)

  smplrt = iim42652['accelODR']
  smpldt = 1 / smplrt

  print('Sample rate', smplrt, 'Hz')

  header = ('Time (s)',)
  if iim42652['powerAccel'] != 'off':
    header += ('Acc_x (g)', 'Acc_y (g)', 'Acc_z (g)')
  if iim42652['powerGyro'] != 'off':
    header += ('Gyr_x (dps)', 'Gyr_y (dps)', 'Gyr_z (dps)')
  if not iim42652['disableTemp'] :
    header += ('Temp. (°C)',)
  if iim42652['powerAccel'] != 'off' and iim42652['powerGyro'] != 'off':
    header += ('Tmst. (ms)',)

  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.output == 'lstofile':
    outf = ls = linesocket.LineSocket()
    lstfcmd = ['python3', './iim42652_spi/lstofile.py',
               '--infile=%s' % (ls.server_address,), '--outfile=%s' % (args.outfile,),
               '--debug=%d' % (args.debug,), '--ilff=%d' % (args.ilff,)]
    print('Launching lstofile helper process: %s' % (' '.join(lstfcmd),))
    p = subprocess.Popen(lstfcmd, shell=False)
    ls.openSocket(1)
  elif args.output == 'socket':
    outf = ls = linesocket.LineSocket()
    ls.openSocket(args.nclients)
  else:
    if args.ilff:
      outf = ilff.ILFFFile(args.outfile, 'w')
    else:
      outf = open(args.outfile, 'w')

  print('Use output:', outf)

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  print('Table(', outf, header, precs, ')')
  outt = table.Table(outf, header, precs)
  outt.writeHeader()

  iloop = 0
  i0 = ilast = 0
  t0 = tlast = tManTempLast = tManMagLast = time.time()

  fcnt, wmfl, ovfl = getFInfo()
  avgfcnt = fcnt
  stdfcnt = 0
  log.msg("FIFO start: count %d, wmfl %d, ovfl %d" % (fcnt, wmfl, ovfl))

  maxfcnt = 0
  nresets = 0
  nwmfl = 0
  avgAlpha = 0.9

  curTempVals = [0]
  blksz = fifoPacketProp.length

  time.sleep(10e-3)
  iim42652['fifoFlush'] = True
  iim42652['fifoMode'] = 'Stream-to-FIFO'

  while True:

    timeval = time.time()
    telapse = timeval - t0
    telapseR = timeval - tlast
    tlast = timeval
    timevalOut = telapse if args.rel_times else timeval

    fcnt, wmfl, ovfl = getFInfo()
    maxfcnt = max(maxfcnt, fcnt)
    avgfcnt = avgfcnt * avgAlpha + fcnt * (1 - avgAlpha)
    stdfcnt = (fcnt - avgfcnt)

    if ovfl:
      fcnt_, wmfl_, ovfl_ = getFInfo()
      log.error(f"FIFO overflow: it {iloop}, FIFO {fcnt} (now: {fcnt_}/{wmfl_}/{ovfl_})")
      nresets += 1
      break
    if wmfl:
      nwmfl += 1
#      log.error(f"FIFO watermark reached: it {iloop}, FIFO {fcnt}")

    if iloop % 1000 == 0:
      sys.stdout.write("%.3f s %.1f µs: %.2f /s " \
                       "FIFO: % 3d (avg %.1f %+.1f, max %d), ovfl %d/%d wm %d/%d: %d B, %.2f KB/s, CSV %d B %.3f MB/s\n" %
                       (telapse, telapseR*1e3, nreads / telapse, fcnt, avgfcnt, stdfcnt, maxfcnt, ovfl, nresets, wmfl, nwmfl,
                        nfifodata, 1e-3 * nfifodata / telapse, outt.nchar, 1e-6 * outt.nchar / telapse ))
      sys.stdout.flush()

    if fcnt <= 8:
      time.sleep(smpldt)
    if process and not ovfl and not wmfl:

      nblk_poss = int(fcnt/blksz)
      nblk = min(40, nblk_poss)

      for i in range(nblk):

        pkg = iim42652[fifoPacketPName]
        # print(pkg)

        nfifodata += blksz

        blktm = timevalOut - (nblk - i - 1) * smpldt
        block = [blktm,] + pkg['values']

        try:
          outt.writeValues(block)
        except BlockingIOError:
          log.msg('socket currently full: FIFO %d' % (fcnt,))
          time.sleep(1e-3)
          break

      nsamples += nblk*blksz
      nreads += nblk

    if args.maxreads >= 0 and nreads > args.maxreads:
      print("\nReached max read samples: %g" % (args.maxreads,))
      break
    if args.maxreadt >= 0 and telapse > args.maxreadt:
      print("\nReached max read time: %g" % (args.maxtime,))
      break
    iloop += 1

  iim42652.enableFIFO = False

  print("read %d samples from %d blocks from queue, wrote %d B CSV data" % (nsamples, nreads, outt.nchar))


if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
