#! /usr/bin/env python3

# python module
import time
import argparse
import math
import struct
import os
import sys
import subprocess
import socket
import fcntl
import termios
import threading
import signal

# fionec modules
import ilff

# package module
from .spitool import SPIConn
from .iim42652 import getIIM42652Cmdline
from . import table
from . import log
from .timer import WTimer
from . import decimate
from . import monitor
from . import senseHatTemperature as st

class Status:
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = []
  nsockdata = 0
  subpid = 0
  subproc = None

class Context:
  done = False
  sock = None
  args = None
  sensor = None
  params = None
  decim = None
  outt = None

status = Status()

def setupSensor():
  iim42652, args = getIIM42652Cmdline()

  iim42652.setupFIFO(args.fifo_packet, args.odr)

  return iim42652, args


def initOutput(context):
  iim42652 = context.sensor
  args = context.args

  if args.temp_ext_enable:
    channels = [st.avail_channels[i] for i in args.temp_ext_enable]
    opts = {'constant_B': args.temp_ext_b}
    st.init(channels, **opts)
    status.curTempVals = st.readall()
  else:
    st.init([])

  header = ('Time (s)',)
  if iim42652['powerAccel'] != 'off':
    aUnit = 'g'
    if args.accel_g != 1:
      aUnit = 'm/s²'
    header += (f'Acceleration_X ({aUnit})', f'Acceleration_Y ({aUnit})', f'Acceleration_Z ({aUnit})')
  if iim42652['powerGyro'] != 'off':
    header += ('Rotation_X (°/s)', 'Rotation_Y (°/s)', 'Rotation_Z (°/s)')
  header += ('Temperature_1 (°C)',)
  if args.tempmanrfreq > 0:
    header += tuple('Temperature_%d (°C)' % (i+2,) for i in args.temp_ext_enable)

  if args.downsample_outfile == '':
    (a, b) = os.path.splitext(args.outfile)
    of100 = a + '_%dHz' % (args.downsample_rate) + b
  else:
    of100 = args.downsample_outfile

  if args.events_outfile == '':
    (a, b) = os.path.splitext(args.outfile)
    eventsof = a + '_events' + b
  else:
    eventsof = args.events_outfile

  if args.ilff:
    outf = ilff.ILFFFile(args.outfile, 'w')
    outf100 = ilff.ILFFFile(of100, 'w')
    log.msg('using ILFF version %s' % (ilff.VERSION))
  else:
    outf = open(args.outfile, 'w')
    outf100 = open(of100, 'w')
  log.msg('Open output mode "%s", file name "%s: %s"' % (args.output, args.outfile, outf))
  log.msg('Open output mode "%s", file name "%s: %s"' % (args.output, of100, outf100))
  log.msg(f'Open output events file {eventsof}')

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  outt = table.Table(outf, header, precs)
  outt.writeHeader()

  tdefFile = 'tabledef.json'
  with open(tdefFile, 'w') as f:
    f.write(repr(outt))

  smplrt = iim42652.fifoSampleRate
  smpldt = 1 / smplrt

  dsFactor = int(round(smplrt / args.downsample_rate))

  log.msg(f'Sample rate: Sensor {iim42652.sampleRate} Hz, FIFO {iim42652.fifoSampleRate} Hz, downsample {args.downsample_rate} Hz (factor {dsFactor})')

  xc_home = '.'
  try:
    xc_home_env = os.getenv('XC_HOME')
    if xc_home_env:
      xc_home = os.path.join(xc_home_env, 'etc')
  except:
    pass

  outt100 = outt.getV3Table(outf100)
  outt100.writeHeader()

  xc_home = '.'
  xc_home_etc = '.'
  try:
    xc_home_env = os.getenv('XC_WORKDIR')
    if xc_home_env:
      xc_home = xc_home_env
      xc_home_etc = os.path.join(xc_home_env, 'etc')
  except:
    pass

  with open(os.path.join(xc_home, 'um-header.csv'), 'w') as f:
    f.write(outt.getHeader())

  monitor.actionsFilename = os.path.join(xc_home_etc, 'actions.json')
  monitor.bordersFilename = os.path.join(xc_home_etc, 'limits.json')
  monitor.eventsFilename = eventsof
  monitor.readHeaders(outt.headers)
  monitor.readBorders()
  monitor.startEventLog()
  monitor.cancelActions()

  outLatest = open(os.path.join(xc_home, 'um-latest.csv'), 'w')

  def writeSlowValues(items):
    ln = outt100.fmtValues(items)
    outLatest.seek(0)
    outLatest.truncate()
    outLatest.write(ln)
    outLatest.flush()
    outt100.writeLine(ln)

  decim = decimate.Decimator(dsFactor, len(header),
                             writeSlowValues,
                             lambda items, mode: monitor.processValues(items, mode))

  log.msg(f'Setup decimation: {decim}')

  context.decim = decim
  context.outt = outt

def processValues(context):
  sock = context.sock
  decim = context.decim
  iim42652 = context.sensor
  outt = context.outt
  args = context.args

  blksz = context.params['blksz']

  smplrt = context.params['smplrt']
  smpldt = 1 / smplrt

  iall = 0
  tbase = 0

  tstamp0 = time.time()

  timer = WTimer()
  if not context.args.rel_times:
    tbase = timer.start

  log.msg('iim42652_fifo_mp queue read loop starting')

  sockdata = b''
  tempManInter = 1e-2

  chnksz = 40960
  if smplrt < 1000:
    chnksz = 4096
  if smplrt < 100:
    chnksz = 64
  if smplrt < 100:
    chnksz = 8

  while not context.done:
    timer.click()

    while True:
      try:
        chunk = sock.recv(chnksz)
#        log.msg(f'Read {chnksz} B from sock, got {len(chunk)} B chunk')
        if len(chunk) == 0:
          context.done = True
          break
        sockdata += chunk
        status.nsockdata += len(sockdata)
        if len(sockdata) > 1e3:
          break
      except BlockingIOError as ex:
        # log.msg('socket queue is empty')
        if len(sockdata):
          break
        else:
          time.sleep(1e-4)
      except KeyboardInterrupt as ex:
        log.msg('Crtl-C')
        break
      except BaseException as ex:
        log.msg('Error reading socket: %s' % (ex,))
        raise(ex)
        break

    # log.msg('Got %d B of data from socket' % (len(sockdata),))
    nblk = int(len(sockdata)/blksz)

    blks = [ sockdata[i*blksz:(i+1)*blksz] for i in range(nblk) ]
    sockdata = sockdata[nblk*blksz:]

    pblks = [ iim42652.FIFOPacketProp.parseData(iim42652.register, blks[i])['values'] for i in range(nblk) ]

    for i in range(nblk):
      block = pblks[i]
      if iim42652.fifoMode > 2:
        block = [tstamp0 + block[-1]*1e-3] + block[0:-1]
      else:
        block = [tstamp0 + (status.nreads + i)*smpldt] + block
      if args.tempmanrfreq > 0:
        block += status.curTempVals

      oln = outt.fmtValues(block)
      outt.writeLine(oln)

      decim.processValues(block)
      iall += 1

    if timer.elapsed('info', 1):
      nread = fcntl.ioctl(sock, termios.FIONREAD, b'\x00' * 4, 1)
      nread = struct.unpack('i', nread)[0]
      log.msg("%s(% 3d): %d B, %.2f KB/s, CSV %d B %.3f MB/s, pending %d B" %
            (">" if 1e-3 * timer.last > smpldt * 40 else " ", 1e-3 * timer.last / smpldt,
             status.nsockdata, 1e-3 * status.nsockdata / timer.elaps,
             outt.nchar, 1e-6 * outt.nchar / timer.elaps, nread))
      sys.stdout.flush()

      if args.maxreads >= 0 and status.nreads > args.maxreads and not context.done:
        print("\nReached max read samples: %g" % (args.maxreads,))
        stopMeasurement()
      if args.maxreadt >= 0 and timer.elaps > args.maxreadt and not context.done:
        print("\nReached max read time: %g" % (args.maxreadt,))
        stopMeasurement()

    status.nsamples += nblk*blksz
    status.nreads += nblk

  log.msg('iim42652_fifo_mp queue read loop finished')
  monitor.stopActions()


def tempReadThreadFun(context, tempmanrfreq=1):
  tempManInter = 1 / tempmanrfreq
  nc = len(st.channels)
  if nc == 0:
    return
  tempManInter = tempManInter / nc
  log.msg('tempReadThreadFun start:', nc, 'channels', tempmanrfreq, 'Hz', tempManInter, 's')
  timer = WTimer()
  ci = 0
  while not context.done:
    timer.click()
    # tlast = timer.elapsInter('tempfetch')
    if status.fcnt < 1000 and tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
      status.curTempVals[ci] = st.read(ci)
      # log.msg('Temp', ci, st.channels, st.ntc_param_B, st.res_R2, status.curTempVals[ci], tlast)
      ci += 1
      ci %= nc
    else:
      time.sleep(1e-3)
  log.msg('tempReadThreadFun exit')

def startProcess(cargs):
  log.msg('starting FIFO reader process')

  args = ['sudo', 'iim42652-fiforeader'] + cargs

  log.msg('start process %s' % (args,))
  proc = subprocess.Popen(args, start_new_session=True)
  log.msg('started process %s,%s' % (args,proc.pid))
  status.subpid = proc.pid
  status.subproc = proc

def stopMeasurement():
  pgid = os.getpgid(0)
  subpgid = os.getpgid(status.subpid)
  os.killpg(subpgid, signal.SIGHUP)

def ctrlCHandler(sig, frame):
  log.msg(f'Signal {sig} caught, stopping helper process')
  stopMeasurement()

def run():
  log.name = 'iim42652_fifo_mp'

  iim42652, args = setupSensor()
  log.msg(iim42652)

  log.level = 2 if args.verbose else 1

  context = Context()
  context.args = args
  context.sensor = iim42652

  log.msg('iim42652_fifo_mp starting')
  log.msg('iim42652_fifo_mp starting helper thread')

  server_address = '/run/usense/linesocket-%d.sock' % (os.getpid(),)

  params = {}
  params['blksz'] = iim42652.getFIFOBlockSize()
  params['smplrt'] = iim42652.fifoSampleRate
  log.msg('params', params)

  context.params = params

  signal.signal(signal.SIGINT, ctrlCHandler)
  signal.signal(signal.SIGTERM, ctrlCHandler)
  signal.signal(signal.SIGHUP, ctrlCHandler)

  initOutput(context)

  tempThread = threading.Thread(target=tempReadThreadFun,args=(context,10))
  log.msg('start temp reader', tempThread)
  tempThread.start()

  startProcess(['--bus', '%d' % args.bus,
                '--dev', '%d' % args.dev,
                '--speed', '%d' % args.speed,
                '--outfile', server_address,
                '--verbose', '%d' % (args.verbose if args.verbose is not None else 0),
                '--fifo-packet', '%d' % args.fifo_packet])

  sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
  sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(1e8))

  t0 = time.time()
  maxWaitTime = 5
  while True:
    log.msg('connecting to {}'.format(server_address))
    try:
      sock.connect(server_address)
      break
    except socket.error as msg:
      failmsg = msg
    if time.time() - t0 > maxWaitTime:
      log.msg('Failed to connect to server in %g s' % (maxWaitTime,))
      log.msg(failmsg)
      sys.exit(1)
    time.sleep(0.5)

  log.msg('connected to {}'.format(server_address))
  sys.stdout.flush()

  sock.setblocking(False)
  context.sock = sock

  log.msg('iim42652_fifo_mp starting main loop')
  try:
    processValues(context)
  except BaseException as ex:
    log.msg(f'Error {ex} occured in main processing loop, exiting program')
    context.done = True
    raise(ex)

  tempThread.join()
  log.msg(f"read {status.nsamples} B, {status.nreads} samples from queue")

if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
