#! /usr/bin/env python3

# python module
import time
import argparse
import math
import struct
import os
import sys
import threading
import queue

# fionec modules
import ilff

# package module
from .iim42652 import getIIM42652Cmdline
from .timer import WTimer
from . import table
from . import log
from . import decimate
from . import monitor
from . import queuereader
from . import socketreader
from . import senseHatTemperature as st

useDecimatorSocket = True

class Status:
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = []

class Context:
  done = False
  que = None
  cmdque = None
  args = None
  sensor = None

status = Status()
context = Context()

def readFIFOSPI(context):
  print('readFIFOSPI start')

  que = context.que
  iim42652 = context.sensor

  while True:
    try:
      cmd = context.cmdque.get()
      if cmd == 'start':
        break
      else:
        print('error: strange start command: %s' % (cmd,))
    except queue.Empty:
      time.sleep(1e-3)
      pass

  print('read FIFO: start')

  iloop = 0
  avgAlpha = 0.8

  timer = WTimer()

  status.fcnt, status.wmfl, status.ovfl = iim42652.getFIFOInfo()

  blksz = iim42652.getFIFOBlockSize()

  iim42652['fifoMode'] = 'Stream-to-FIFO'
  print('readFIFOSPI start II')

  maxReadBlkSize = 1024

  def getQueue(fetch=0):
    fcnt, wmfl, ovfl = iim42652.getFIFOInfo()
    timer.click()
    bcnt = int(min(fcnt,maxReadBlkSize)/blksz)
    if bcnt > 0:
      # values = [ iim42652.getFIFOPacketValues() for i in range(bcnt) ]
      fifoblob = iim42652.getFIFOPacketData(bcnt)
      fifodblocks = [ fifoblob[i*blksz:(i+1)*blksz] for i in range(bcnt) ]
      status.nfifodata += bcnt*blksz
    else:
      # values = []
      fifodblocks = []

    status.maxfcnt = max(status.maxfcnt, fcnt)
    tstmp = timer.elaps if context.args.rel_times else timer.time

    if ovfl:
      iim42652.resetFIFO()
      log.error("FIFO overflow: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, bcnt))
      status.nresets += 1
      sys.stdout.flush()
    elif wmfl:
      status.nwmfl += 1
      #log.error("FIFO watermark reached: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, len(fifodata)))

    status.fcnt, status.wmfl, status.ovfl = (fcnt, wmfl, ovfl)
    return (fifodblocks, tstmp)

  smplrt = iim42652.fifoSampleRate
  smpldt = 1 / smplrt

  while True:
    iloop += 1

    while True:
      (fifodata, tmstmp) = getQueue()
      if len(fifodata) > 0:
        que.put((tmstmp, fifodata), block=False)
      if len(fifodata) < 250:
        break

    if status.fcnt == 0:
      if context.done:
        break
      else:
        if not status.ovfl and not status.wmfl:
          time.sleep(smpldt/2)

#    if timer.elapsed('break', 150) and False:
#      print('break')
#      time.sleep(0.1)

    if status.fcnt < 200:
      time.sleep(2e-3)

    if que.qsize() > 10000:
      log.error('queue got too large (%d), stopping' % (que.qsize()))
      break

  print('readFIFOSPI stop')

def processValues(context):
  iim42652 = context.sensor
  args = context.args
  que = context.que

  channels = [st.avail_channels[i] for i in args.temp_ext_enable]
  opts = {'constant_B': args.temp_ext_b}
  st.init(channels, **opts)
  status.curTempVals = st.readall()

  print('Compose header')
  print('powerAccel:', iim42652['powerAccel'])
  print('powerGyro:', iim42652['powerGyro'])
  
  header = ('Time (s)',)
  if iim42652['powerAccel'] != 'off':
    unit = 'g' if args.accel_g else 'm/s²'
    header += ('Acceleration_X (%s)' % (unit,), 'Acceleration_Y (%s)' % (unit,), 'Acceleration_Z (%s)' % (unit,))
  if iim42652['powerGyro'] != 'off':
    header += ('Rotation_X (°/s)', 'Rotation_Y (°/s)', 'Rotation_Z (°/s)')
  header += ('Temperature_1 (°C)',)
  if args.tempmanrfreq > 0:
    header += tuple('Temperature_%d (°C)' % (i+1,) for i in args.temp_ext_enable)

  log.msg(f'Table header {header}')
    
  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.ilff:
    outf = ilff.ILFFFile(args.outfile, 'w')
    print('using ILFF version %s' % (ilff.VERSION))
  else:
    outf = open(args.outfile, 'w')

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  outt = table.Table(outf, header, precs)

  tdefFile = 'tabledef.json'
  with open(tdefFile, 'w') as f:
    f.write(repr(outt))

  if useDecimatorSocket:
    outf.close()
  else:
    outt.writeHeader()
    
  if args.downsample_outfile == '':
    (a, b) = os.path.splitext(args.outfile)
    of100 = a + '_%dHz' % (args.downsample_rate) + b
  else:
    of100 = args.downsample_outfile

  smplrt = iim42652.fifoSampleRate
  smpldt = 1 / smplrt

  print('Sample rate: %g' % (smplrt,))
  print('dt         : %g' % (1 / smplrt,))

  dsFactor = int(round(smplrt / args.downsample_rate))

  xc_home = '.'
  xc_home_etc = '.'
  try:
    xc_home_env = os.getenv('XC_WORKDIR')
    if xc_home_env:
      xc_home = xc_home_env
      xc_home_etc = os.path.join(xc_home_env, 'etc')
  except:
    pass

  if useDecimatorSocket:
    subargs = ['--outfile', args.outfile]
    subargs += ['--outfile-slow', of100]
    subargs += ['--output', 'ilff' if args.ilff else args.output]
    subargs += ['--downsample-factor', '%d' % (dsFactor,)]
    subargs += ['--actions-file', os.path.join(xc_home_etc, 'actions.json')]
    subargs += ['--limits-file', os.path.join(xc_home_etc, 'limits.json')]
    subargs += ['--event-file', os.path.join(xc_home, 'events.csv')]
    subargs += ['--header-file', os.path.join(xc_home, 'um-header.csv')]
    subargs += ['--latest-file', os.path.join(xc_home, 'um-latest.csv')]
    subargs += ['--table-def-file', tdefFile]
    socketReader = socketreader.SocketReader()
    socketReader.startProcess(subargs)

  else:
    outf100 = open(of100, 'w')
    outt100 = outt.getV3Table(outf100)
    outt100.writeHeader()

    monitor.actionsFilename = 'actions.json'
    monitor.bordersFilename = 'limits.json'
    monitor.readHeaders(header)
    monitor.readBorders()
    monitor.cancelActions()
    monitor.eventsFilename = os.path.join(xc_home, 'events.csv')
    monitor.startEventLog()

    decim = decimate.Decimator(dsFactor, len(header),
                               lambda items: outt100.writeValues(items),
                               lambda items, mode: monitor.processValues(items, mode))

    qreaderDecimateAndActions = queuereader.QueueReaderMT(lambda x: decim.processValues(x))

  timer = WTimer()
  log.msg('iim42652_fifo_mt queue read loop started and ready')

  context.cmdque.put('start')

  tstamp0 = None
  nblkout = 0

  tempManInter = 10
  if args.tempmanrfreq > 0:
    tempManInter = 1 / args.tempmanrfreq

  while True:
    timer.click()

    fifodblocks = []
    try:
      (tstamp, fifodblocks) = que.get(False)
      if tstamp0 is None:
        tstamp0 = tstamp
      # time.sleep(1e-3)
    except queue.Empty:
      if context.done:
        print('Queue is empty and stop set')
        break
      else:
        time.sleep(smpldt/2)

    nblk = len(fifodblocks)
    #print('Got %d B = %d blocks' % (len(fifodata), nblk))
    block = []

    for i in range(nblk):
      dblock = fifodblocks[i]
      block = iim42652.FIFOPacketProp.parseData(iim42652.register, dblock)['values']
      if iim42652.fifoMode > 2:
        block = [tstamp0 + block[-1]*1e-3] + block[0:-1]
      else:
        block = [tstamp0 + (status.nreads + i)*smpldt] + block
      if args.tempmanrfreq > 0:
        block += status.curTempVals

      if useDecimatorSocket:
        oln = outt.fmtValues(block)
        for i in range(100):
          try:
            socketReader.putLine(oln)
            break
          except BlockingIOError as ex:
            print('socked blocked', ex)
            if i == 99:
              print('socked blocked definitely', ex)
              raise(ex)
            else:
              time.sleep(i*1e-3)
          except BaseException as ex:
            print(ex)
            raise(ex)
      else:
        outt.writeValues(block)
        qreaderDecimateAndActions.que.put(block)

    status.nsamples += nblk*len(block)
    status.nreads += nblk

    if args.tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
      status.curTempVals = st.readall()

    if timer.elapsed('info', 1):
      print("%s%s: FIFO: %d (max %d), ov %d wm %d: %d B, %.2f kB/s, CSV %d B %.3f MB/s, Queue %d" %
                       (">" if timer.last > 40 else " ", timer, status.fcnt, status.maxfcnt,
                        status.nresets, status.nwmfl,
                        status.nfifodata, 1e-3 * status.nfifodata / timer.elaps,
                        outt.nchar, 1e-6 * outt.nchar / timer.elaps, que.qsize() ))
      sys.stdout.flush()
      if status.fcnt == 0:
        if timer.elapsed('infoempty', 10):
          print('FIFO is currently empty')

    if args.maxreads >= 0 and status.nreads > args.maxreads and not context.done:
      print("\nReached max read samples: %g" % (args.maxreads,))
      iim42652['fifoModeInt'] = 0
      context.done = True
    if args.maxreadt >= 0 and timer.elaps > args.maxreadt and not context.done:
      print("\nReached max read time: %g" % (args.maxreadt,))
      iim42652['fifoModeInt'] = 0
      context.done = True

  if not useDecimatorSocket:
    qreaderDecimateAndActions.done = context.done
  log.msg('iim42652_fifo_mt queue read loop finished')

def setupSensor():
  iim42652, args = getIIM42652Cmdline()
  iim42652.setupFIFO(args.fifo_packet, args.odr)
  return iim42652, args

def run():
  log.name = 'iim42652_fifo'

  iim42652, args = setupSensor()
  print(iim42652)

  context.que = queue.SimpleQueue()
  context.cmdque = queue.SimpleQueue()
  context.args = args
  context.sensor = iim42652

  log.msg('iim42652_fifo_mt starting')
  log.msg('iim42652_fifo_mt starting helper thread')

  rFIFOThread = threading.Thread(target=readFIFOSPI, args=(context,))
  rFIFOThread.start()

  log.msg('iim42652_fifo_mt starting main loop')
  processValues(context)

  assert(context.done == True)

  rFIFOThread.join()

  print("read %d samples from %d blocks from queue" % (status.nsamples, status.nreads))

if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
