#! /usr/bin/env python3

import time
import argparse
import math
import struct
import os
import sys
import subprocess

import ilff

from .iim42652 import IIM42652
from . import spitool
from . import table
from . import linesocket
from . import log

regWhoami = 0
valWhoami = 0xEA

regReset = 0x06
valReset = 0x41

def parseargs():
  parser = argparse.ArgumentParser(description='Erzeuge Umspul-Konfiguration.')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(10e6), help='SPI speed')

  parser.add_argument('--debug', action='store_true', help='Trace SPI')
  parser.add_argument('--trace', action='store_true', help='Trace SPI')
  parser.add_argument('--reset', action='store_true', help='Perform initial reset')
  parser.add_argument('--noreset', action='store_true', help='No initial reset')

  parser.add_argument('--ilff', action='store_true', help='Trace SPI')
  parser.add_argument('--output', type=str, default="ilff", help='Trace SPI')
  parser.add_argument('--outfile', '-o', type=str, default="out.csv", help='Trace SPI')

  parser.add_argument('--precision', '-p', type=int, default=7, help='CSV precision')
  parser.add_argument('--format', type=str, help='CSV value format (fixed or adaptive)')
  parser.add_argument('--rel-times', action='store_true', help='Reform initial reset')

  parser.add_argument('--maxreads', type=int, default=-1, help='number of reads')
  parser.add_argument('--maxtime', type=float, default=-1, help='time limit of reading')

  parser.add_argument('--odr', '-r', type=int, default=500, help='Trace SPI')

  args = parser.parse_args()
  return args


def run():
  log.name = 'iim42652_fifo'
  import locale
  locale.setlocale(locale.LC_ALL, '')

  args = parseargs()
  if args.trace:
    print(args)
    print(f'locale: {locale.getlocale()}')
    IIM42652.DEBUG = True
    spitool.DEBUG = True

  spi = spitool.SPIConn(args.bus, args.dev, args.speed)

  iim42652 = IIM42652(spi)
  if not iim42652.connected:
    print("device not found")
    return

  if not args.noreset or args.reset:
    iim42652.reset()

  nfifodata = 0

  fifoInfoPropName = 'FIFO_INFO_BLOB'
  fifoInfoProp = iim42652.getProperty(fifoInfoPropName)

  def getFInfo():
    data = iim42652[fifoInfoPropName]
    f = fifoInfoProp.splitr(None, data)
    return ( f[v] for v in ['fi_FIFO_COUNT', 'fi_fifoThreshInt', 'fi_fifoFullInt'] )

  rawWords = []
  nreads = 0
  nsamples = 0
  process = True

  iim42652['accelODR'] = args.odr
  smplrt = iim42652['accelODR']
  smpldt = 1 / smplrt

  print('Sample rate', smplrt, 'Hz')
  
  iim42652['powerAccel'] = 'lownoise'

  header = ('Time (s)',)
  if iim42652['powerAccel'] != 'off':
    header += ('Acc_x (g)', 'Acc_y (g)', 'Acc_z (g)')
  if iim42652['powerGyro'] != 'off':
    header += ('Gyr_x (dps)', 'Gyr_y (dps)', 'Gyr_z (dps)')
  if not iim42652['disableTemp'] :
    header += ('Temp. (°C)',)

  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.output == 'lstofile':
    outf = ls = linesocket.LineSocket()
    lstfcmd = ['python3', './iim42652_spi/lstofile.py',
               '--infile=%s' % (ls.server_address,), '--outfile=%s' % (args.outfile,),
               '--debug=%d' % (args.debug,), '--ilff=%d' % (args.ilff,)]
    print('Launching lstofile helper process: %s' % (' '.join(lstfcmd),))
    p = subprocess.Popen(lstfcmd, shell=False)
    ls.openSocket(1)
  elif args.output == 'socket':
    outf = ls = linesocket.LineSocket()
    ls.openSocket(args.nclients)
  elif args.output == 'ilff':
    outf = ilff.ILFFFile(args.outfile, 'w')
  else:
    outf = open(args.outfile, 'w')

  print('Use output:', outf)

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  print('Table(', outf, header, precs, ')')
  outt = table.Table(outf, header, precs)
  outt.writeHeader()

  iloop = 0
  i0 = ilast = 0
  t0 = tlast = tManTempLast = tManMagLast = time.time()

  iim42652['fifoMode'] = 'Stream-to-FIFO'
  iim42652['FIFO_WM'] = 1000
  iim42652['FIFO_CONFIG1'] = 0
  iim42652['fifoAccelEn'] = True
  iim42652['fifoTempEn'] = True
  iim42652['fifoWmGtTh'] = True
  iim42652['fifoResumePartialRd'] = True

  fcnt, wmfl, ovfl = getFInfo()
  avgfcnt = fcnt
  stdfcnt = 0
  log.msg("FIFO start: count %d, wmfl %d, ovfl %d" % (fcnt, wmfl, ovfl))

  maxfcnt = 0
  nresets = 0
  nwmfl = 0
  avgAlpha = 0.9

  curTempVals = [0]
  fPktProp = getattr(IIM42652.Register, 'fifoPkgFmt1')
  blksz = fPktProp.length

  while True:

    timeval = time.time()
    telapse = timeval - t0
    telapseR = timeval - tlast
    tlast = timeval
    timevalOut = telapse if args.rel_times else timeval

    fcnt, wmfl, ovfl = getFInfo()
    maxfcnt = max(maxfcnt, fcnt)
    avgfcnt = avgfcnt * avgAlpha + fcnt * (1 - avgAlpha)
    stdfcnt = (fcnt - avgfcnt)

    if ovfl:
      fcnt_, wmfl_, ovfl_ = getFInfo()
      log.error(f"FIFO overflow: it {iloop}, FIFO {fcnt} (now: {fcnt_}/{wmfl_}/{ovfl_})")
      nresets += 1
      break
    if wmfl:
      nwmfl += 1
#      log.error(f"FIFO watermark reached: it {iloop}, FIFO {fcnt}")

    if iloop % 1000 == 0:
      sys.stdout.write("%.5f %.5f: FIFO: %d (avg %.1f %+.1f, max %d), ovfl %d/%d wm %d/%d: %d B, %.2f KB/s, CSV %d B %.3f MB/s\n" %
                       (telapse, telapseR, fcnt, avgfcnt, stdfcnt, maxfcnt, ovfl, nresets, wmfl, nwmfl,
                        nfifodata, 1e-3 * nfifodata / telapse, outt.nchar, 1e-6 * outt.nchar / telapse ))
      sys.stdout.flush()

    if fcnt <= 8:
      time.sleep(smpldt)
    if process and not ovfl and not wmfl:

      nblk_poss = int(fcnt/blksz)
      nblk = min(40, nblk_poss)

      dummy = iim42652.spi.getbytes(fPktProp.source, nblk * blksz)

      nsamples += nblk*blksz
      nreads += nblk

    if args.maxreads >= 0 and nreads > args.maxreads:
      print("\nReached max read samples: %g" % (args.maxreads,))
      break
    if args.maxtime >= 0 and telapse > args.maxtime:
      print("\nReached max read time: %g" % (args.maxtime,))
      break
    iloop += 1

  iim42652.enableFIFO = False

  print("read %d samples from %d blocks from queue, wrote %d B CSV data" % (nsamples, nreads, outt.nchar))


if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
