#! /usr/bin/env python3
# -*- coding:utf-8 -*-

import time
import math
import re
import os
import json

import ilff

from . import log
from . import actions
from . import alerts
from . import timer

DEBUG = False

borders = {'min': {}, 'max': {}, 'wmin': {}, 'wmax': {}, 'warnlev': {}}
actionstore = {'warnl': {}, 'alarml': {}, 'warnh': {}, 'alarmh': {}}

headers = []
units = []

bordersFilename = ''
stlast = None

actionsFilename = ''
afile = None

eventsFilename = ''

normAccel = 9.80665

def readBorders(fname=None):
    if fname is None:
        fname = bordersFilename
    log.msg('Read borders from %s' % (fname,))

    global stlast
    stlast = os.stat(fname)

    data = json.loads(open(fname, 'r').read())

    limits = ['min', 'wmin', 'max', 'wmax']

    for name in data:
        for l in limits:
            try:
                val = float(data[name][l])
            except:
                if 'min' in l:
                    val = -math.inf
                else:
                    val = math.inf
                print(f'Error: failed to read number {data[name][l]} for limit {l} of {name}, set to {val}')

            rval = val
            unit = None
            if name in headers:
                hi = headers.index(name)
                unit = units[hi]
            if unit == 'm/s²':
                val *= normAccel
            if DEBUG:
                print(f'Read border {l} for {name}: unit={unit}, raw value {rval}, value={val}')

            borders[l][name] = val

        actionstore['alarmh'][name] = 'redlight'
        actionstore['warnh'][name] = 'orangelight'
        actionstore['warnl'][name] = 'cyanlight'
        actionstore['alarml'][name] = 'bluelight'

        # actionstore['alarmh'][name] = 'redlightBlink'
        # actionstore['warnh'][name] = 'redlight'
        # actionstore['warnl'][name] = 'bluelight'
        # actionstore['alarml'][name] = 'bluelightBlink'

def readHeaders(line):
    global afile
    afile = open(actionsFilename, 'w')

    if isinstance(line, str):
        if line[0] == '#':
            line = line[1:]
        items = line.split(';')
    else:
        items = line

    alerts.initModule()
    alerts.setValueInfo(items)

    items = [f.strip() for f in items]

    itemunits = [f for f in map(lambda x: re.search(r'([^ ]+) \((.+)\)', x).group(2), items)]
    items = [f for f in map(lambda x: re.search(r'([^ ]+).*', x).group(1), items)]

    global headers, units
    headers = items
    units = itemunits

def runAction(a):
    if a in dir(actions):
        f = getattr(actions, a)
        f()
        ## log.msg('Run action %s' % (a,))
    else:
        log.msg('Error: Action %s not found' % (a,))

def cancelAction(a):
    actions.resetactions()

def cancelActions():
    actions.resetactions()

def stopActions():
    actions.stopactions()

def extractSignals(actions):
    types = [f['type'] for f in actions]
    if len(types) > 0:
        minlev = min(types)
        maxlev = max(types)
        if abs(minlev) > abs(maxlev):
            mlev = minlev
        else:
            mlev = maxlev
        where = types.index(mlev)
        sig = [actions[where]]
        return sig
    else:
        return []

class EventLog:
    eventfile = '/var/lib/usense/events.csv'

    def __init__(self, eventfile=None):
        if eventfile:
            self.eventfile = eventfile
        self.file = ilff.ILFFFile(self.eventfile, 'w')

    def addEvent(self, data):
        ln = '%.3f; %s' % (time.time(), json.dumps(data))
        # print('event', ln)
        self.file.write(ln)

    def flush(self):
        self.file.flush()

def startEventLog():
    global eventLog
    eventLog = EventLog(eventsFilename)

def runActions(actions, oldActions):
    afile.truncate()
    afile.seek(0)
    afile.write(json.dumps(actions))
    afile.flush()

    signals = extractSignals(actions)
    oldSignals = extractSignals(oldActions)

    signals = [f['action'] for f in signals]
    oldSignals = [f['action'] for f in oldSignals]

    if signals != oldSignals:
        #        print('actions', actions)
        #        print('signals', signals)
        eventLog.addEvent(actions)
        for (i, a) in enumerate(signals):
            runAction(a)
        if len(actions) == 0:
            cancelActions()

    if len(signals) > 0:
        alerts.send_email_alert(actions)

    # if len(actions) == 0:
    #     cancelActions()

warned = {}
status = {'lastActions': {'min': [], 'max': [], 'minmax': []}}
timer = timer.WTimer()

def processValues(items, mode):

    timer.click()
    if timer.elapsed('bodercheck', 1):
        # print('check limits file stat')
        st = os.stat(bordersFilename)
        if st is None:
            print('error: File not found: %s' % (bordersFilename,))
        elif st.st_mtime > stlast.st_mtime:
            readBorders(bordersFilename)
        eventLog.flush()

    states = [(0,0) for f in items]

    # items is concatenation of 100Hz min and 100Hz max values
    assert(len(items) == 2*len(headers))

    for (i, val) in enumerate(items):
        head = headers[i % len(headers)]

        # log.msg(head, borders['min'])

        if not head in borders['min']:
            if head not in warned:
                print('Header %s not found in borders' % (head,))
                print('Borders: %s' % (borders['min'].keys(),))
                warned[head] = 1
            continue

        bmin = borders['min'][head]
        bmax = borders['max'][head]
        bminw = borders['wmin'][head]
        bmaxw = borders['wmax'][head]

    #            log.msg(i, head, bmin, bmax)
        state = 0
        score = val
        if val < bmin:
            state = -2
            score = -val
        elif val < bminw:
            state = -1
            score = -val
        elif val > bmax:
            state = 2
        elif val > bmaxw:
            state = 1

        # log.msg('Value %s = %g: %d\n' % (head, val, state,))

        states[i] = (state, score)

    actlist = []
    #        log.msg(states)

    for (i, s) in enumerate(states):
        score = s[1]
        state = s[0]
        head = headers[i % len(headers)]
        act = None
    #            log.msg(head)
    #            log.msg(s)
    #            log.msg(actions['warn'])
        if state == 0:
            pass
        elif state == 1:
            act = actionstore['warnh'][head]
        elif state == 2:
            act = actionstore['alarmh'][head]
        elif state == -1:
            act = actionstore['warnl'][head]
        elif state == -2:
            act = actionstore['alarml'][head]
        actlist += [{'action': act, 'type': state, 'sev': abs(state), 'value': items[i], 'score': score, 'source': head}]

    # sort according to (source, sev, score)
    actlist = sorted(actlist, key=lambda v: v['score'], reverse=True)
    actlist = sorted(actlist, key=lambda v: abs(v['sev']), reverse=True)
    actlist = sorted(actlist, key=lambda v: v['source'])

    # every second selected: worst one of min and max part of list
    actlist = actlist[::2]

    actlist = [{'action': a['action'], 'type': a['type'], 'value': a['value'], 'source': a['source']}
                 for a in actlist if a['action'] is not None]

    runActions(actlist, status['lastActions'][mode])
    status['lastActions'][mode] = actlist


if __name__ == "__main__":
    log.name = 'filtervals'

    bname = 'Grenzen.ini'

    statedir = '/run/usense/%d' % (os.getuid(),)
    sname = '%s/um-values.sock' % (statedir,)

    if len(sys.argv) > 1:
        bname = sys.argv[1]

    if len(sys.argv) > 2:
        sname = sys.argv[2]
        statedir = os.path.dirname(sname)

    aname = '%s/um-actions.json' % (statedir,)
    if len(sys.argv) > 3:
        aname = sys.argv[3]
    afile = open(aname, 'w')

    bordersFilename = bname
    readBorders(bname)
    readValues(sname, processValuesDecimate)
