#! /usr/bin/env python3

import queue
import threading
import time
import subprocess
import argparse
import sys
import socket
import shutil

import ilff

from . import linesocket
from . import table
from . import decimate
from . import monitor
from . import log

class SocketReader:

    def __init__(self):
        self.que = queue.Queue()
        self.done = False
        self.sleepInter = 2e-3
        srmain = shutil.which('iim42652-socketreader')
        print('socketreader main:', srmain)
        if srmain:
            self.mainProgram = srmain
        else:
            self.mainProgram = ['./iim42652_spi/socketreader.py']

    def readLoop(self, server_address):
        print('socket reader thread started')

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(1e8))


        while True:
            log.msg('connecting to {}'.format(server_address))
            try:
                sock.connect(server_address)
                break
            except socket.error as msg:
                failmsg = msg
            if time.time() - t0 > maxWaitTime:
                log.msg('Failed to connect to server in %g s' % (maxWaitTime,))
                log.msg(failmsg)
                sys.exit(1)
            time.sleep(0.5)

        log.msg('connected to {}'.format(server_address))
        sys.stdout.flush()

        sockfd = sock.makefile()

        while True:
            ln = sockfd.readline()
            if len(ln) == 0:
                print('socket closed')
                break

            self.process(ln)

        print('socket reader thread stopped')
        monitor.stopActions()

    def startProcess(self, cargs):
        print('starting socket reader process')

        self.ls = linesocket.LineSocket()

        args = [self.mainProgram]
        args += ['--infile', self.ls.server_address]
        args += cargs

        print('start process %s' % (args,))

        proc = subprocess.Popen(args)

        self.ls.openSocket(1)

    def put(self, items):
        self.ls.write(';'.join(['%.16g' % (v,) for v in items]) + '\n')

    def putLine(self, vline):
        self.ls.write('%s\n' % (vline,))

    def initOutput(self, args):

        log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
        self.ofname = args.outfile
        self.ofname_slow = args.outfile_slow

        if args.output == 'ilff':
            self.outf = ilff.ILFFFile(self.ofname, 'w')
            self.outf_slow = ilff.ILFFFile(self.ofname_slow, 'w')
        else:
            self.outf = open(self.ofname, 'w')
            self.outf_slow = open(self.ofname_slow, 'w')

        self.outt = table.Table(self.outf, [], [])

        with open(args.table_def_file, 'r') as f:
            jsdata = f.read()
            self.outt.readJSON(jsdata)

        self.outt.writeHeader()

        with open(args.header_file, 'w') as f:
            f.write(self.outt.getHeader())

        self.outLatest = open(args.latest_file, 'w')

        self.outt100 = self.outt.getV3Table(self.outf_slow)
        self.outt100.writeHeader()

        monitor.actionsFilename = args.actions_file
        monitor.bordersFilename = args.limits_file
        monitor.eventsFilename = args.event_file
        monitor.readHeaders(self.outt.headers)
        monitor.readBorders()
        monitor.startEventLog()
        monitor.cancelActions()

        self.decim = decimate.Decimator(args.downsample_factor, len(self.outt.headers),
                                        lambda items: self.writeSlowValues(items),
                                        lambda items, mode: monitor.processValues(items, mode))

    def writeSlowValues(self, items):
        ln = self.outt100.fmtValues(items)
        self.outLatest.seek(0)
        self.outLatest.truncate()
        self.outLatest.write(ln)
        self.outLatest.flush()

        self.outt100.writeLine(ln)
    def process(self, ln):
        self.outt.write(ln)
        parts = ln.split(';')
        items = [float(v) for v in parts]
        self.decim.processValues(items)

def run():
    log.name = 'socketreader'

    parser = argparse.ArgumentParser(description='Socket reader helper process')

    parser.add_argument('--infile', type=str, help='name of input file')
    parser.add_argument('--outfile', type=str, help='name of output file')
    parser.add_argument('--outfile-slow', type=str, help='name of (slow) output file')

    parser.add_argument('--downsample-factor', dest='downsample_factor', type=int, help='factor for downsampling')

    parser.add_argument('--output', type=str, help='')

    parser.add_argument('--limits-file', type=str, help='')
    parser.add_argument('--actions-file', type=str, help='')
    parser.add_argument('--table-def-file', type=str, help='')
    parser.add_argument('--event-file', type=str, help='')
    parser.add_argument('--header-file', type=str, help='')
    parser.add_argument('--latest-file', type=str, help='')

    args = parser.parse_args(sys.argv[1:])

    print(args)

    sockreader = SocketReader()
    sockreader.initOutput(args)
    sockreader.readLoop(args.infile)

if __name__ == "__main__":
    run()
